# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python Hangman game."""

from absl import flags
from absl.testing import absltest
import numpy as np

from open_spiel.python.games import hangman
import pyspiel

_WORD_LIST_FILE = flags.DEFINE_string("word_list_file", "", "Word list file")

_SEED = 329827811
_NUM_SIMULATIONS = 100


class HangmanTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(_SEED)
    params = {"word_list_file": _WORD_LIST_FILE.value}
    self.game = pyspiel.load_game("python_hangman", params)

  def run_random_simulation(self, word: str | None = None):
    if word is not None:
      # Start with a specific word.
      state = self.game.new_initial_state(word)
    else:
      # Randomly sample a word via an initial chance node.
      state = self.game.new_initial_state()
      self.assertEqual(state.current_player(), pyspiel.PlayerId.CHANCE)
      outcomes = state.chance_outcomes()
      action_list, prob_list = zip(*outcomes)
      action = np.random.choice(action_list, p=prob_list)
      print("Sampled outcome: ",
            state.action_to_string(state.current_player(), action))
      state.apply_action(action)
    print(state)
    self.assertEqual(state.current_player(), 0)
    self.assertLen(state.legal_actions(), hangman.NUM_DISTINCT_ACTIONS)
    while not state.is_terminal():
      action = np.random.choice(state.legal_actions())
      action_string = state.action_to_string(state.current_player(), action)
      print(f"Player {state.current_player()} " +
            f"randomly sampled action: {action_string}")
      state.apply_action(action)
      print(f"Rewards: {state.rewards()}")
      print(state)
    print("Terminal state:")
    print(state)
    print(f"Rewards: {state.rewards()}")
    print(f"Returns: {state.returns()}")

  def test_random_simulations_chance_node(self):
    """Tests basic API functions."""
    for g in range(_NUM_SIMULATIONS):
      print("")
      print(f"Starting game: {g}")
      self.run_random_simulation()

  def test_random_simulations_specific_words(self):
    for word in ["pineapple", "does", "not", "belong", "on", "pizza"]:
      print("")
      print(f"Starting game with fixed word: {word}")
      self.run_random_simulation(word)

  def test_word_with_spaces(self):
    state = self.game.new_initial_state("hello world")
    self.assertEqual(state.current_player(), 0)
    action_sequence = [ord("h") - 97, ord("e") - 97, ord("l") - 97,
                       ord("o") - 97, ord("w") - 97, ord("r") - 97,
                       ord("d") - 97]
    for action in action_sequence:
      print("")
      print(state)
      print(f"Guessing letter: {chr(action + 97)}")
      state.apply_action(action)
      self.assertGreater(state.rewards()[0], 0.0)
    print("Final state:")
    print(state)
    self.assertTrue(state.is_terminal())
    self.assertAlmostEqual(state.returns()[0],
                           7*hangman.CORRECT_GUESS_REWARD + hangman.WIN_REWARD)

  def test_all_incorrect_guesses(self):
    state = self.game.new_initial_state("banana")
    self.assertEqual(state.current_player(), 0)
    action_sequence = [ord("h") - 97, ord("e") - 97, ord("l") - 97,
                       ord("o") - 97, ord("w") - 97, ord("r") - 97]
    for action in action_sequence:
      self.assertFalse(state.is_terminal())
      print("")
      print(state)
      print(f"Guessing letter: {chr(action + 97)}")
      state.apply_action(action)
      self.assertLessEqual(state.rewards()[0], 0.0)
    self.assertTrue(state.is_terminal())
    print("Final state:")
    print(state)
    self.assertAlmostEqual(state.returns()[0], self.game.min_utility())


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from absl.testing import absltest
import numpy as np

import pyspiel


class GamesRepeatedPokerTest(absltest.TestCase):

  def test_bindings(self):
    acpc_game_string = (
        "universal_poker(betting=nolimit,numPlayers=2,numRounds=4,blind=100 50,"
        "firstPlayer=2 1 1 1,numSuits=4,numRanks=13,numHoleCards=2,"
        "numBoardCards=0 3 1 1,stack=20000 20000)"
    )
    game_string = (
        f"repeated_poker(universal_poker_game_string={acpc_game_string},"
        "max_num_hands=3,reset_stacks=True,rotate_dealer=True)"
    )
    game = pyspiel.load_game(game_string)
    state = game.new_initial_state()

    self.assertEqual(state.dealer(), 1)
    self.assertEqual(state.small_blind(), 50)
    self.assertEqual(state.big_blind(), 100)
    self.assertEqual(state.stacks(), [20000, 20000])
    self.assertEqual(state.player_to_seat(0), 0)
    self.assertEqual(state.player_to_seat(1), 1)
    self.assertEqual(state.seat_to_player(0), 0)
    self.assertEqual(state.seat_to_player(1), 1)
    self.assertEqual(state.dealer_seat(), 1)
    self.assertEqual(state.small_blind_seat(), 1)
    self.assertEqual(state.big_blind_seat(), 0)
    self.assertEqual(state.acpc_hand_histories(), [])

    rng = np.random.RandomState(0)
    while not state.is_terminal():
      if state.is_chance_node():
        outcomes = state.chance_outcomes()
        action_list, prob_list = zip(*outcomes)
        action = rng.choice(action_list, p=prob_list)
        state.apply_action(action)
      else:
        state.apply_action(rng.choice(state.legal_actions()))
    self.assertLen(state.acpc_hand_histories(), 3)
    for history in state.acpc_hand_histories():
      self.assertIsInstance(history, str)

  def test_state_struct(self):
    acpc_game_string = pyspiel.hunl_game_string("fullgame")
    game_string = (
        f"repeated_poker(universal_poker_game_string={acpc_game_string},"
        "max_num_hands=100,reset_stacks=True,rotate_dealer=True)")
    game = pyspiel.load_game(game_string)
    state = game.new_initial_state()
    state_struct = state.to_struct()
    self.assertEqual(state_struct.hand_number, 0)
    self.assertEqual(state_struct.max_num_hands, 100)
    self.assertEqual(state_struct.stacks, [20000, 20000])
    self.assertEqual(state_struct.dealer, 1)
    self.assertEqual(state_struct.small_blind, 50)
    self.assertEqual(state_struct.big_blind, 100)
    self.assertLen(state_struct.hand_returns, 1)
    self.assertEqual(state_struct.hand_returns[0], [0.0, 0.0])
    up_json = json.loads(state_struct.current_universal_poker_json)
    self.assertEqual(up_json["current_player"], pyspiel.PlayerId.CHANCE)
    self.assertEqual(state_struct.prev_universal_poker_json, "")
    state.apply_action(state.string_to_action("player=-1 move=Deal 2c"))
    state.apply_action(state.string_to_action("player=-1 move=Deal 2d"))
    state.apply_action(state.string_to_action("player=-1 move=Deal Tc"))
    state.apply_action(state.string_to_action("player=-1 move=Deal 2s"))
    state.apply_action(state.string_to_action("player=1 move=Bet300"))
    state.apply_action(state.string_to_action("player=0 move=Fold"))
    self.assertEqual(state.returns(), [0.0, 0.0])
    state_struct = state.to_struct()
    prev_up_json = json.loads(state_struct.prev_universal_poker_json)
    self.assertEqual(prev_up_json["player_hands"], ["2d2c", "Tc2s"])


if __name__ == "__main__":
  if "universal_poker" in pyspiel.registered_names():
    absltest.main()
  else:
    print("universal_poker is not registered. Skipping test.")



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General tests for pyspiel python bindings."""

import json
import os

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python import games  # pylint: disable=unused-import
from open_spiel.python import policy
from open_spiel.python.mfg import games as mfgs  # pylint: disable=unused-import
import pyspiel


_FULLY_OPTIONAL_PYTHON_GAMES = frozenset([
    "python_pokerkit_wrapper",
    "python_pokerkit_wrapper_acpc_style",
    "python_repeated_pokerkit"
])
# Specify game names in alphabetical order, to make the test easier to read.
# "Mandatory" = neither optional nor included only if certain flags are set.
EXPECTED_MANDATORY_GAMES = frozenset([
    "2048",
    "add_noise",
    "amazons",
    "antichess",
    "backgammon",
    "bargaining",
    "battleship",
    "blackjack",
    "blotto",
    "breakthrough",
    "bridge",
    "bridge_uncontested_bidding",
    "cached_tree",
    "catch",
    "chat_game",  # python game locating in python/games/chat_games/
    "checkers",
    "chess",
    "cliff_walking",
    "clobber",
    "coin_game",
    "colored_trails",
    "connect_four",
    "coop_box_pushing",
    "coop_to_1p",
    "coordinated_mp",
    "crazy_eights",
    "crazyhouse",
    "cribbage",
    "cursor_go",
    "dark_chess",
    "dark_hex",
    "dark_hex_ir",
    "deep_sea",
    "dots_and_boxes",
    "dou_dizhu",
    "efg_game",
    "einstein_wurfelt_nicht",
    "euchre",
    "first_sealed_auction",
    "gin_rummy",
    "go",
    "gomoku",
    "goofspiel",
    "havannah",
    "hex",
    "hearts",
    "hive",
    "kriegspiel",
    "kuhn_poker",
    "laser_tag",
    "latent_ttt",
    "lewis_signaling",
    "leduc_poker",
    "liars_dice",
    "liars_dice_ir",
    "lines_of_action",
    "maedn",
    "mancala",
    "markov_soccer",
    "matching_pennies_3p",
    "matrix_bos",
    "matrix_brps",
    "matrix_cd",
    "matrix_coordination",
    "matrix_mp",
    "matrix_pd",
    "matrix_rps",
    "matrix_rpsw",
    "matrix_sh",
    "matrix_shapleys_game",
    "mean_field_lin_quad",
    "mfg_crowd_modelling",
    "mfg_crowd_modelling_2d",
    "mfg_dynamic_routing",
    "mfg_garnet",
    "misere",
    "mnk",
    "morpion_solitaire",
    "negotiation",
    "nfg_game",
    "nim",
    "nine_mens_morris",
    "normal_form_extensive_game",
    "oh_hell",
    "oshi_zumo",
    "othello",
    "oware",
    "pentago",
    "pathfinding",
    "phantom_go",
    "phantom_ttt",
    "phantom_ttt_ir",
    "pig",
    "python_ant_foraging",
    "python_block_dominoes",
    "python_dynamic_routing",
    "python_hangman",
    "python_iterated_prisoners_dilemma",
    "python_mfg_crowd_avoidance",
    "python_mfg_crowd_modelling",
    "python_mfg_dynamic_routing",
    "python_mfg_periodic_aversion",
    "python_mfg_predator_prey",
    "python_kuhn_poker",
    "python_team_dominoes",
    "python_tic_tac_toe",
    "python_liars_poker",
    "quoridor",
    "repeated_game",
    "repeated_leduc_poker",
    "rbc",
    "restricted_nash_response",
    "sheriff",
    "skat",
    "start_at",
    "solitaire",
    "spades",
    "stones_and_gems",
    "tarok",
    "tic_tac_toe",
    "tiny_bridge_2p",
    "tiny_bridge_4p",
    "tiny_hanabi",
    "trade_comm",
    "turn_based_simultaneous_game",
    "twixt",
    "ultimate_tic_tac_toe",
    "y",
    "yacht",
    "zerosum",
])


class PyspielTest(parameterized.TestCase):

  def test_registered_names_is_sorted(self):
    game_names = pyspiel.registered_names()
    self.assertSequenceEqual(game_names, sorted(game_names))

  def test_registered_names_contains_expected_games(self):
    game_names = pyspiel.registered_names()

    expected = list(EXPECTED_MANDATORY_GAMES)
    if (os.environ.get("OPEN_SPIEL_BUILD_WITH_HANABI", "OFF") == "ON" and
        "hanabi" not in expected):
      expected.append("hanabi")
    if (os.environ.get("OPEN_SPIEL_BUILD_WITH_ACPC", "OFF") == "ON" and
        "universal_poker" not in expected):
      expected.append("universal_poker")
      expected.append("repeated_poker")
    # Verify we have registered all mandatory games + games added via flags
    self.assertContainsSubset(expected, game_names)

    # Check that the contents of the remainder are all fully optional games.
    # (If this fails, it likely means that someone added a game but forgot to
    # update either EXPECTED_MANDATORY_GAMES or _FULLY_OPTIONAL_PYTHON_GAMES.)
    remaining_games = set(game_names).difference(set(expected))
    self.assertContainsSubset(expected_subset=remaining_games,
                              actual_set=_FULLY_OPTIONAL_PYTHON_GAMES)

  def test_default_loadable(self):
    # Games which cannmot be loaded with default parameters will be skipped by
    # several standard tests. We make a list of such games here in order to make
    # implementors think twice about making new games non-default-loadable
    non_default_loadable = [
        game.short_name
        for game in pyspiel.registered_games()
        if not game.default_loadable
    ]
    expected = [
        # Being non-default-loadable prevents various automated tests.
        # Only add games here if there is no sensible default for a parameter.
        "add_noise",
        "cached_tree",
        "coop_to_1p",
        "efg_game",
        "nfg_game",
        "misere",
        "turn_based_simultaneous_game",
        "normal_form_extensive_game",
        "repeated_game",
        "restricted_nash_response",
        "start_at",
        "zerosum",
    ]
    if (os.environ.get("OPEN_SPIEL_BUILD_WITH_ACPC", "OFF") == "ON" and
        "repeated_poker" not in expected):
      expected.append("repeated_poker")
    self.assertCountEqual(non_default_loadable, expected)

  def test_registered_game_attributes(self):
    game_list = {game.short_name: game for game in pyspiel.registered_games()}
    self.assertEqual(game_list["kuhn_poker"].dynamics,
                     pyspiel.GameType.Dynamics.SEQUENTIAL)
    self.assertEqual(game_list["kuhn_poker"].chance_mode,
                     pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC)
    self.assertEqual(game_list["kuhn_poker"].information,
                     pyspiel.GameType.Information.IMPERFECT_INFORMATION)
    self.assertEqual(game_list["kuhn_poker"].utility,
                     pyspiel.GameType.Utility.ZERO_SUM)
    self.assertEqual(game_list["kuhn_poker"].min_num_players, 2)

  def test_create_game(self):
    game = pyspiel.load_game("kuhn_poker")
    game_info = game.get_type()
    self.assertEqual(game_info.information,
                     pyspiel.GameType.Information.IMPERFECT_INFORMATION)
    self.assertEqual(game.num_players(), 2)

  def test_play_kuhn_poker(self):
    game = pyspiel.load_game("kuhn_poker")
    state = game.new_initial_state()
    self.assertEqual(state.is_chance_node(), True)
    self.assertEqual(state.chance_outcomes(), [(0, 1 / 3), (1, 1 / 3),
                                               (2, 1 / 3)])
    state.apply_action(1)
    self.assertEqual(state.is_chance_node(), True)
    self.assertEqual(state.chance_outcomes(), [(0, 0.5), (2, 0.5)])
    state.apply_action(2)
    self.assertEqual(state.is_chance_node(), False)
    self.assertEqual(state.legal_actions(), [0, 1])
    sampler = pyspiel.UniformProbabilitySampler(0., 1.)
    clone = state.resample_from_infostate(1, sampler)
    self.assertEqual(
        clone.information_state_string(1), state.information_state_string(1))

  def test_othello(self):
    game = pyspiel.load_game("othello")
    state = game.new_initial_state()
    self.assertFalse(state.is_chance_node())
    self.assertFalse(state.is_terminal())
    self.assertEqual(state.legal_actions(), [19, 26, 37, 44])

  def test_tic_tac_toe(self):
    game = pyspiel.load_game("tic_tac_toe")
    state = game.new_initial_state()
    self.assertFalse(state.is_chance_node())
    self.assertFalse(state.is_terminal())
    self.assertEqual(state.legal_actions(), [0, 1, 2, 3, 4, 5, 6, 7, 8])

  def test_game_string(self):
    game = pyspiel.load_game("tic_tac_toe")
    self.assertEqual(str(game), game.to_string())

  def test_game_parameters_from_string_empty(self):
    self.assertEqual(pyspiel.game_parameters_from_string(""), {})

  def test_game_parameters_from_string_simple(self):
    self.assertEqual(
        pyspiel.game_parameters_from_string("foo"), {"name": "foo"})

  def test_game_parameters_from_string_with_options(self):
    self.assertEqual(
        pyspiel.game_parameters_from_string("foo(x=2,y=true)"), {
            "name": "foo",
            "x": 2,
            "y": True
        })

  def test_game_parameters_from_string_with_subgame(self):
    self.assertEqual(
        pyspiel.game_parameters_from_string(
            "foo(x=2,y=true,subgame=bar(z=False))"), {
                "name": "foo",
                "x": 2,
                "y": True,
                "subgame": {
                    "name": "bar",
                    "z": False
                }
            })

  def test_game_parameters_to_string_empty(self):
    self.assertEqual(pyspiel.game_parameters_to_string({}), "")

  def test_game_parameters_to_string_simple(self):
    self.assertEqual(
        pyspiel.game_parameters_to_string({"name": "foo"}), "foo()")

  def test_game_parameters_to_string_with_options(self):
    self.assertEqual(
        pyspiel.game_parameters_to_string({
            "name": "foo",
            "x": 2,
            "y": True
        }), "foo(x=2,y=True)")

  def test_game_parameters_to_string_with_subgame(self):
    self.assertEqual(
        pyspiel.game_parameters_to_string({
            "name": "foo",
            "x": 2,
            "y": True,
            "subgame": {
                "name": "bar",
                "z": False
            }
        }), "foo(subgame=bar(z=False),x=2,y=True)")

  def test_game_type(self):
    game_type = pyspiel.GameType(
        "matrix_mp", "Matching Pennies", pyspiel.GameType.Dynamics.SIMULTANEOUS,
        pyspiel.GameType.ChanceMode.DETERMINISTIC,
        pyspiel.GameType.Information.PERFECT_INFORMATION,
        pyspiel.GameType.Utility.ZERO_SUM,
        pyspiel.GameType.RewardModel.TERMINAL, 2, 2, True, True, False, False,
        dict())
    self.assertEqual(game_type.chance_mode,
                     pyspiel.GameType.ChanceMode.DETERMINISTIC)

  def test_error_handling(self):
    with self.assertRaisesRegex(RuntimeError,
                                "Unknown game 'invalid_game_name'"):
      unused_game = pyspiel.load_game("invalid_game_name")

  def test_can_create_cpp_tabular_policy(self):
    for game_name in ["kuhn_poker", "leduc_poker", "liars_dice"]:
      game = pyspiel.load_game(game_name)

      # We just test that we can create a tabular policy.
      policy.python_policy_to_pyspiel_policy(policy.TabularPolicy(game))

  def test_simultaneous_game_history(self):
    game = pyspiel.load_game("coop_box_pushing")
    state = game.new_initial_state()
    state.apply_action(0)
    state2 = game.new_initial_state()
    state2.apply_actions([0] * game.num_players())
    self.assertEqual(state.history(), state2.history())

  def test_record_batched_trajectories(self):
    for game_name in ["kuhn_poker", "leduc_poker", "liars_dice"]:
      game = pyspiel.load_game(game_name)
      python_policy = policy.TabularPolicy(game)
      tabular_policy = policy.python_policy_to_pyspiel_policy(python_policy)
      policies = [tabular_policy] * 2

      # We test that we can create a batch of trajectories.
      seed = 0
      batch_size = 128
      include_full_observations = False
      pyspiel.record_batched_trajectories(game, policies,
                                          python_policy.state_lookup,
                                          batch_size, include_full_observations,
                                          seed, -1)

  def test_get_game_parameters(self):
    if "universal_poker" in pyspiel.registered_names():
      game = pyspiel.load_game(pyspiel.hunl_game_string("fullgame"))
      print(game)
      params = game.get_parameters()
      print(params)
      param_string = pyspiel.game_parameters_to_string(params)
      print(param_string)
      params2 = pyspiel.game_parameters_from_string(param_string)
      assert params2 is not None
      game_from_params = pyspiel.load_game(
          f"{game.get_type().short_name}{param_string}")
      self.assertGreaterEqual(len(params2), len(params))
      assert game_from_params is not None

  @parameterized.parameters(
      ("universal_poker", pyspiel.hunl_game_string("fullgame")),
      (
          "repeated_poker",
          "repeated_poker(max_num_hands=10,reset_stacks=True,"
          + "rotate_dealer=True,universal_poker_game_string=universal_poker("
          + "betting=nolimit,bettingAbstraction=fullgame,blind=100 50,"
          + "firstPlayer=2 1 1 1,numBoardCards=0 3 1 1,numHoleCards=2,"
          + "numPlayers=2,numRanks=13,numRounds=4,numSuits=4,stack=20000 20000"
          + "))"),
      ("python_pokerkit_wrapper", "python_pokerkit_wrapper()"),
      ("python_pokerkit_wrapper",
       "python_pokerkit_wrapper(variant=FixedLimitSevenCardStud)"),
      ("python_pokerkit_wrapper",
       "python_pokerkit_wrapper(variant=PotLimitOmahaHoldem)"),
      ("python_pokerkit_wrapper_acpc_style",
       "python_pokerkit_wrapper_acpc_style(),"),
      ("kuhn_poker", "kuhn_poker(players=3)"),
      ("tic_tac_toe", "tic_tac_toe"),
      ("breakthrough", "breakthrough(rows=6,columns=6)"))
  def test_game_serialize_deserialize(self, game_name, game_string):
    if game_name in pyspiel.registered_names():
      game = pyspiel.load_game(game_string)
      serialized_game = game.serialize()
      game2 = pyspiel.deserialize_game(serialized_game)
      self.assertEqual(str(game), str(game2))

  def test_game_structs(self):
    game = pyspiel.load_game("tic_tac_toe")
    state = game.new_initial_state()

    # Test ActionStruct
    action = state.legal_actions()[0]
    action_struct = state.action_to_struct(action)
    self.assertIsNotNone(action_struct)
    self.assertIsInstance(action_struct, pyspiel.SpielStruct)
    self.assertIsInstance(action_struct, pyspiel.ActionStruct)
    json_str = action_struct.to_json()
    self.assertIn('"row":', json_str)
    self.assertIn('"col":', json_str)

    # Test StructToActions
    self.assertEqual([action], state.struct_to_actions(action_struct))

    # Test StateStruct
    state_struct = state.to_struct()
    self.assertIsNotNone(state_struct)
    self.assertIsInstance(state_struct, pyspiel.SpielStruct)
    self.assertIsInstance(state_struct, pyspiel.StateStruct)
    json_str = state_struct.to_json()
    self.assertIn('"board":', json_str)

    # Test ObservationStruct
    observation_struct = state.to_observation_struct()
    self.assertIsNotNone(observation_struct)
    self.assertIsInstance(observation_struct, pyspiel.SpielStruct)
    self.assertIsInstance(observation_struct, pyspiel.ObservationStruct)
    json_str = observation_struct.to_json()
    self.assertIn('"board":', json_str)


class StructApiTest(absltest.TestCase):
  """Tests for struct API."""

  def setUp(self):
    super().setUp()
    self.game = pyspiel.load_game("tic_tac_toe")
    self.ttt = pyspiel.tic_tac_toe
    # Create a reference state with one move applied
    self.ref_state = self.game.new_initial_state()
    self.ref_state.apply_action(0)  # X plays top-left

  # ===== Struct Field Access Tests =====

  def test_state_struct_fields(self):
    """Test that struct fields are accessible and correct."""
    state_struct = self.ref_state.to_struct()
    self.assertEqual(state_struct.current_player, "o")
    self.assertEqual(state_struct.board,
                     ["x", ".", ".", ".", ".", ".", ".", ".", "."])

  def test_observation_struct_fields(self):
    """Test observation struct has correct fields."""
    obs_struct = self.ref_state.to_observation_struct(0)
    self.assertEqual(obs_struct.current_player, "o")
    self.assertEqual(obs_struct.board,
                     ["x", ".", ".", ".", ".", ".", ".", ".", "."])

  def test_action_struct_fields(self):
    """Test action struct has semantic row/col fields via JSON."""
    action = 4  # Center square
    action_struct = self.ref_state.action_to_struct(action)
    # Note: TicTacToeActionStruct fields aren't directly bound in Python yet,
    # but we can verify via JSON
    parsed = json.loads(action_struct.to_json())
    self.assertEqual(parsed["row"], 1)
    self.assertEqual(parsed["col"], 1)

  # ===== JSON Conversion Tests =====

  def test_to_json(self):
    """Test state converts to expected JSON string."""
    state_struct = self.ref_state.to_struct()
    json_str = state_struct.to_json()
    self.assertIn('"current_player":', json_str)
    self.assertIn('"board":', json_str)
    # Verify it's valid JSON
    parsed = json.loads(json_str)
    self.assertEqual(parsed["current_player"], "o")

  def test_struct_from_json(self):
    """Test struct can be constructed from JSON string."""
    json_str = self.ref_state.to_json()
    state_struct = self.ttt.TicTacToeStateStruct(json_str)
    self.assertEqual(state_struct.to_json(), json_str)

  def test_json_round_trip(self):
    """Test JSON -> struct -> JSON preserves data."""
    original_json = self.ref_state.to_json()
    state_struct = self.ttt.TicTacToeStateStruct(original_json)
    round_trip_json = state_struct.to_json()
    self.assertEqual(original_json, round_trip_json)

  # ===== Dict Conversion Tests =====

  def test_to_dict(self):
    """Test state converts to Python dict."""
    state_dict = self.ref_state.to_dict()
    self.assertIsInstance(state_dict, dict)
    self.assertIn("current_player", state_dict)
    self.assertIn("board", state_dict)
    self.assertEqual(state_dict["current_player"], "o")

  def test_state_from_dict(self):
    """Test creating state from dict."""
    state_dict = {
        "board": ["x", "o", ".", ".", ".", ".", ".", ".", "."],
        "current_player": "x",
    }
    state = self.game.new_initial_state(state_dict)
    self.assertEqual(state.to_dict(), state_dict)

  def test_state_to_from_dict_round_trip(self):
    """Test state -> dict -> state round-trip."""
    state = self.game.new_initial_state()
    for action in [0, 1, 4]:
      state.apply_action(action)

    state_dict = state.to_dict()
    new_state = self.game.new_initial_state(state_dict)
    self.assertEqual(new_state.to_dict(), state_dict)

  # ===== Starting State Tests =====

  def test_empty_starting_state(self):
    """Test that normal initial state has no custom starting state."""
    state = self.game.new_initial_state()
    state.apply_action(0)
    self.assertIsNone(state.starting_state())
    self.assertEqual(state.starting_state_str(), "")

  def test_starting_state_from_dict(self):
    """Test creating state from dict sets starting_state correctly."""
    state = self.game.new_initial_state(self.ref_state.to_dict())
    self.assertEmpty(state.history())
    self.assertIsNotNone(state.starting_state())
    self.assertEqual(str(state.starting_state()), str(self.ref_state))

  def test_starting_state_str(self):
    """Test starting_state_str returns JSON."""
    state = self.game.new_initial_state(self.ref_state.to_dict())
    starting_str = state.starting_state_str()
    self.assertNotEqual(starting_str, "")
    # Should be valid JSON
    parsed = json.loads(starting_str)
    self.assertEqual(parsed["current_player"], "o")

  def test_starting_state_unchanged_after_action(self):
    """Test that starting_state doesn't change when actions are applied."""
    state = self.game.new_initial_state(self.ref_state.to_dict())
    original_starting = str(state.starting_state())
    state.apply_action(4)  # Apply an action
    self.assertEqual(str(state.starting_state()), original_starting)

  # ===== Action Struct Tests =====

  def test_action_struct_round_trip(self):
    """Test action -> struct -> actions preserves the action."""
    state = self.game.new_initial_state()
    for action in state.legal_actions():
      action_struct = state.action_to_struct(action)
      recovered_actions = state.struct_to_actions(action_struct)
      self.assertEqual([action], recovered_actions)

  def test_apply_action_struct(self):
    """Test applying action via struct produces same result as direct apply."""
    action = self.ref_state.legal_actions()[0]
    action_struct = self.ref_state.action_to_struct(action)

    state1 = self.game.new_initial_state()
    state1.apply_action(0)
    state1.apply_action(action)

    state2 = self.game.new_initial_state()
    state2.apply_action(0)
    state2.apply_action_struct(action_struct)  # Should succeed (no exception)

    self.assertEqual(str(state1), str(state2))

  def test_validate_action_struct_valid(self):
    """Test validate_action_struct succeeds for valid action."""
    action = self.ref_state.legal_actions()[0]
    action_struct = self.ref_state.action_to_struct(action)
    self.ref_state.validate_action_struct(action_struct)  # Should not raise

  def test_validate_action_struct_invalid(self):
    """Test validate_action_struct raises error for invalid action."""
    state = self.game.new_initial_state()
    center_action = 4
    action_struct = state.action_to_struct(center_action)

    state.apply_action(center_action)

    status = state.validate_action_struct(action_struct)
    self.assertFalse(status.ok())

  def test_apply_action_struct_invalid(self):
    """Test apply_action_struct raises error and doesn't mutate state."""
    state = self.game.new_initial_state()
    center_action = 4
    action_struct = state.action_to_struct(center_action)

    state.apply_action(center_action)
    state_str_before = str(state)

    status = state.apply_action_struct(action_struct)
    self.assertFalse(status.ok())

    self.assertEqual(str(state), state_str_before)

  def test_action_struct_from_json(self):
    """Test action struct can be constructed from JSON."""
    action = 4  # Center
    action_struct = self.ref_state.action_to_struct(action)
    json_str = action_struct.to_json()

    # Reconstruct and verify
    parsed = json.loads(json_str)
    self.assertEqual(parsed["row"], 1)
    self.assertEqual(parsed["col"], 1)

  def test_action_struct_row_col_mapping(self):
    """Test action struct row/col correctly maps to board position."""
    state = self.game.new_initial_state()
    for action in range(9):
      action_struct = state.action_to_struct(action)
      parsed = json.loads(action_struct.to_json())
      expected_row = action // 3
      expected_col = action % 3
      self.assertEqual(parsed["row"], expected_row,
                       f"Action {action} should have row {expected_row}")
      self.assertEqual(parsed["col"], expected_col,
                       f"Action {action} should have col {expected_col}")

  # ===== Vector Action API Tests =====

  def test_struct_to_actions(self):
    """Test struct_to_actions returns list with single action."""
    action = 4
    action_struct = self.ref_state.action_to_struct(action)
    actions = self.ref_state.struct_to_actions(action_struct)
    self.assertIsInstance(actions, list)
    self.assertLen(actions, 1)
    self.assertEqual(actions[0], action)

  def test_actions_to_struct(self):
    """Test actions_to_struct with single action."""
    action = 4
    action_struct = self.ref_state.actions_to_struct([action])
    recovered = self.ref_state.struct_to_actions(action_struct)
    self.assertEqual(recovered, [action])

  def test_actions_to_struct_with_player(self):
    """Test actions_to_struct with player parameter."""
    action = 4
    action_struct = self.ref_state.actions_to_struct(1, [action])
    parsed = json.loads(action_struct.to_json())
    self.assertEqual(parsed["row"], 1)
    self.assertEqual(parsed["col"], 1)

  # ===== Constructor Tests =====

  def test_struct_default_constructor(self):
    """Test struct default constructor."""
    s = self.ttt.TicTacToeStateStruct()
    self.assertEqual(s.current_player, "")
    self.assertEqual(s.board, [])

  def test_struct_json_constructor(self):
    """Test struct JSON string constructor."""
    json_str = (
        '{"current_player": "x", "board": [".", ".", ".", ".", ".", ".", ".",'
        ' ".", "."]}'
    )
    s = self.ttt.TicTacToeStateStruct(json_str)
    self.assertEqual(s.current_player, "x")
    self.assertLen(s.board, 9)

  def test_struct_dict_constructor(self):
    """Test struct dict constructor."""
    d = {
        "current_player": "o",
        "board": ["x", ".", ".", ".", ".", ".", ".", ".", "."]
    }
    s = self.ttt.TicTacToeStateStruct(d)
    self.assertEqual(s.current_player, "o")
    self.assertEqual(s.board[0], "x")


class LoadGameFromJsonTest(absltest.TestCase):
  """Tests for load_game_from_json API."""

  def test_load_game_basic(self):
    """Test basic game loading from JSON."""
    game = pyspiel.load_game_from_json('{"game_name": "tic_tac_toe"}')
    self.assertEqual(game.get_type().short_name, "tic_tac_toe")

  def test_load_game_with_params(self):
    """Test game loading with parameters."""
    game = pyspiel.load_game_from_json(
        '{"game_name": "connect_four", "columns": 8}')
    self.assertEqual(game.get_type().short_name, "connect_four")

  def test_load_game_invalid_name(self):
    """Test that invalid game name raises error."""
    with self.assertRaises(pyspiel.SpielError):
      pyspiel.load_game_from_json('{"game_name": "nonexistent_game"}')

  def test_load_game_missing_name(self):
    """Test that missing game_name raises error."""
    with self.assertRaises(pyspiel.SpielError):
      pyspiel.load_game_from_json('{"some_param": 123}')


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Boltzmann DQN agent implemented in JAX.

This algorithm is a variation of DQN that uses a softmax policy directly with
the unregularized action-value function. See https://arxiv.org/abs/2102.01585.
"""

import copy
from typing import Any

import numpy as np
import torch
from torch import nn
import torch.nn.functional as F

from open_spiel.python.pytorch import dqn


class BoltzmannDQN(dqn.DQN):
  """Boltzmann DQN implementation in JAX."""

  def __init__(self, *args, eta: float = 1.0, **kwargs):
    """Initializes the Boltzmann DQN agent.

    Args:
      *args: args passed to the underlying DQN agent.
      eta: Temperature parameter used in the softmax function.
      **kwargs: kwargs passed to the underlying DQN agent.
    """
    super().__init__(*args, **kwargs)
    self._prev_q_network = copy.deepcopy(self._q_network)
    self._temperature = eta

  def _boltzmann_action_probs(
      self,
      network: nn.Module,
      info_state: list[Any],
      legal_actions: list[Any],
      coeff: np.ndarray = None,
  ):
    """Returns a valid softmax action and action probabilities.

    Args:
      network: (nn.Module) Parameters of the Q-network.
      info_state: (list) Observations from the environment.
      legal_actions (list): List of legal actions.
      coeff (np.ndarray): If not None, then the terms in softmax function will
        be element-wise multiplied with these coefficients.

    Returns:
      A valid softmax action and action probabilities.
    """
    info_state = torch.FloatTensor(
        np.asarray(info_state).reshape(1, -1), device=self._device
    )
    q_values = network(info_state).detach().squeeze(0)

    illegal_actions_mask = torch.logical_not(
        torch.BoolTensor(legal_actions, device=self._device)
    )
    legal_q_values = q_values.masked_fill(
        illegal_actions_mask, dqn.ILLEGAL_ACTION_LOGITS_PENALTY
    )
    # Apply temperature and subtract the maximum value for numerical stability.
    temp = legal_q_values / self._temperature
    probs = F.softmax(coeff * temp, -1).detach().cpu().numpy()
    action = np.random.choice(np.arange(self._num_actions), p=probs)
    return action, probs

  def _act_epsilon_greedy(
      self,
      info_state, legal_actions, epsilon: float
  ):
    """Returns a selected action and the probabilities of legal actions."""
    if epsilon == 0:  # greedy evaluation mode
      # Soft-max normalized by the action probabilities from the previous
      # Q-network.
      _, prev_probs = self._boltzmann_action_probs(
          self._prev_q_network,
          info_state,
          legal_actions,
          torch.ones(self._num_actions, device=self._device),
      )
      return self._boltzmann_action_probs(
          self._q_network,
          info_state,
          legal_actions,
          torch.as_tensor(prev_probs, device=self._device),
      )

    # During training, we use the DQN action selection, which will be
    # epsilon-greedy.
    return super()._act_epsilon_greedy(info_state, legal_actions, epsilon)

  def update_prev_q_network(self) -> None:
    """Updates the parameters of the previous Q-network."""
    self._prev_q_network.load_state_dict(
        copy.deepcopy(self._q_network.state_dict()))



#!/usr/bin/env python3
#
# Copyright 2025 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This script generates kCRC32CPowers[]."""


def poly_mul(a, b):
  """Polynomial multiplication: a * b."""
  product = 0
  for i in range(b.bit_length()):
    if (b & (1 << i)) != 0:
      product ^= a << i
  return product


def poly_div(a, b):
  """Polynomial division: floor(a / b)."""
  q = 0
  while a.bit_length() >= b.bit_length():
    q ^= 1 << (a.bit_length() - b.bit_length())
    a ^= b << (a.bit_length() - b.bit_length())
  return q


def poly_reduce(a, b):
  """Polynomial reduction: a mod b."""
  return a ^ poly_mul(poly_div(a, b), b)


def poly_exp(a, b, g):
  """Polynomial exponentiation: a^b mod g."""
  if b == 1:
    return poly_reduce(a, g)
  c = poly_exp(a, b // 2, g)
  c = poly_mul(c, c)
  if b % 2 != 0:
    c = poly_mul(c, a)
  return poly_reduce(c, g)


def bitreflect(a, num_bits):
  """Reflects the bits of the given integer."""
  if a.bit_length() > num_bits:
    raise ValueError(f'Integer has more than {num_bits} bits')
  return sum(((a >> i) & 1) << (num_bits - 1 - i) for i in range(num_bits))


G = 0x11EDC6F41  # The CRC-32C reducing polynomial, in the "natural" bit order
CRC_BITS = 32  # The degree of G, i.e. the 32 in "CRC-32C"
LSB_FIRST = True  # CRC-32C is a least-significant-bit-first CRC
NUM_SIZE_BITS = 64  # The maximum number of bits in the length (size_t)
NUM_DROPPED_BITS = 4  # The number of bits dropped from the length
LOG2_BITS_PER_BYTE = 3  # log2 of the number of bits in a byte, i.e. log2(8)
X = 2  # The polynomial 'x', in the "natural" bit order


def print_crc32c_powers():
  """Generates kCRC32CPowers[].

  kCRC32CPowers[] is an array of length NUM_SIZE_BITS - NUM_DROPPED_BITS,
  whose i'th entry is x^(2^(i + LOG2_BITS_PER_BYTE + NUM_DROPPED_BITS) -
  CRC_BITS - 1) mod G. See kCRC32CPowers[] in the C++ source for more info.
  """
  for i in range(NUM_SIZE_BITS - NUM_DROPPED_BITS):
    poly = poly_exp(
        X,
        2 ** (i + LOG2_BITS_PER_BYTE + NUM_DROPPED_BITS)
        - CRC_BITS
        - (1 if LSB_FIRST else 0),
        G,
    )
    poly = bitreflect(poly, CRC_BITS)
    print(f'0x{poly:0{2*CRC_BITS//8}x}, ', end='')


if __name__ == '__main__':
  print_crc32c_powers()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl import app
from absl import logging
from absl.testing import absltest
from absl.testing import parameterized
import torch

from open_spiel.python import policy
import pyspiel
from open_spiel.python.pytorch import deep_cfr

SEED = 24984617


class DeepCFRPyTorchTest(parameterized.TestCase):

  @parameterized.parameters('leduc_poker', 'kuhn_poker', 'liars_dice')
  def test_deep_cfr_runs(self, game_name):
    game = pyspiel.load_game(game_name)
    deep_cfr_solver = deep_cfr.DeepCFRSolver(
        game,
        policy_network_layers=(8, 4),
        advantage_network_layers=(4, 2),
        num_iterations=2,
        num_traversals=2,
        learning_rate=1e-3,
        batch_size_advantage=None,
        batch_size_strategy=None,
        memory_capacity=1e7,
        seed=SEED,
    )
    deep_cfr_solver.solve()

  def test_matching_pennies_3p(self):
    game = pyspiel.load_game_as_turn_based('matching_pennies_3p')
    deep_cfr_solver = deep_cfr.DeepCFRSolver(
        game,
        policy_network_layers=(16, 8),
        advantage_network_layers=(32, 16),
        num_iterations=2,
        num_traversals=2,
        learning_rate=1e-3,
        batch_size_advantage=None,
        batch_size_strategy=None,
        memory_capacity=int(1e7),
    )
    deep_cfr_solver.solve()
    conv = pyspiel.nash_conv(
        game,
        policy.python_policy_to_pyspiel_policy(
            policy.tabular_policy_from_callable(
                game, deep_cfr_solver.action_probabilities)))
    logging.info('Deep CFR in Matching Pennies 3p. NashConv: %.2f', conv)


def main(_):
  torch.manual_seed(SEED)
  absltest.main()


if __name__ == '__main__':
  # Necessary to run main via app.run for internal tests.
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python Deep CFR example."""

# pylint: disable=logging-fstring-interpolation

from absl import app
from absl import flags
from absl import logging

from open_spiel.python import policy
from open_spiel.python.algorithms import expected_game_score
from open_spiel.python.algorithms import exploitability
from open_spiel.python.jax import deep_cfr
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_integer("num_iterations", 101, "Number of iterations")
flags.DEFINE_integer("num_traversals", 375, "Number of traversals/games")
flags.DEFINE_string("game_name", "kuhn_poker", "Name of the game")

# Recommended parameters:
#   For more, see https://github.com/aicenter/openspiel_reproductions/

#   Parameter                        Value
#   ---------------------------------------
#   num_traversals                   1500
#   batch_size_advantage             2048
#   batch_size_strategy              2048
#   num_hidden                       64
#   num_layers                       3
#   reinitialize_advantage_networks  True
#   learning_rate                    1e-3
#   memory_capacity                  1e6
#   policy_network_train_steps       5000
#   advantage_network_train_steps    750


def main(unused_argv):
  logging.info("Loading %s", FLAGS.game_name)

  game = pyspiel.load_game(FLAGS.game_name)
  deep_cfr_solver = deep_cfr.DeepCFRSolver(
      game,
      policy_network_layers=(64,),
      advantage_network_layers=(64,),
      num_iterations=FLAGS.num_iterations,
      num_traversals=FLAGS.num_traversals,
      reinitialize_advantage_networks=True,
      learning_rate=1e-3,
      batch_size_advantage=256,
      batch_size_strategy=256,
      memory_capacity=100000,
      policy_network_train_steps=2500,
      advantage_network_train_steps=375,
      print_nash_convs=False,  # for debugging purposes
  )

  _, advantage_losses, policy_loss = deep_cfr_solver.solve()
  for player, losses in advantage_losses.items():
    logging.info("Advantage for player %d: %s", player,
                 losses[:2] + ["..."] + losses[-2:])
    logging.info(
        f"Advantage Buffer Size for player {player}:"
        f" {len(deep_cfr_solver.advantage_buffers[player])}"
    )
  logging.info(f"Strategy Buffer Size: {len(deep_cfr_solver.strategy_buffer)}")
  logging.info(f"Final policy loss: {policy_loss}")

  average_policy = policy.tabular_policy_from_callable(
      game, deep_cfr_solver.action_probabilities)

  conv = exploitability.nash_conv(game, average_policy)
  logging.info(f"Deep CFR in {FLAGS.game_name} - NashConv: {conv}")

  average_policy_values = expected_game_score.policy_value(
      game.new_initial_state(), [average_policy] * 2)
  if FLAGS.game_name == "kuhn_poker":
    # We know EVs
    logging.info(
        f"Computed player 0 value: {average_policy_values[0]:.2f} (expected:"
        f" {-1/18:.2f})."
    )
    logging.info(
        f"Computed player 1 value: {average_policy_values[1]:.2f} (expected:"
        f" {1/18:.2f})."
    )
  else:
    logging.info(f"Computed player 0 value: {average_policy_values[0]:.2f}")
    logging.info(f"Computed player 1 value: {average_policy_values[1]:.2f}")


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Export the model's graph."""

import json
import os

from absl import app
from absl import flags
from flax import linen
from flax import nnx

from open_spiel.python.algorithms.alpha_zero import utils
import pyspiel


FLAGS = flags.FLAGS
flags.DEFINE_string("game", None, "Name of the game")

flags.DEFINE_string("save_path", None, "Directory to save graph")
flags.DEFINE_string(
    "path", None, "A path (directory) for a pretrained model chekpoint"
)
flags.DEFINE_integer(
    "checkpoint_step", None, "A step for a pretrained model chekpoint"
)
flags.DEFINE_string(
    "config_path", None, "Filename for the a training config.json"
)


flags.DEFINE_enum(
    "nn_model",
    "resnet",
    utils.api_selector(utils.AVIALABLE_APIS[0]).Model.valid_model_types,
    "What type of model should be used?.",
)
flags.DEFINE_integer("nn_width", 2**7, "How wide should the network be.")
flags.DEFINE_integer("nn_depth", 10, "How deep should the network be.")
flags.DEFINE_float("learning_rate", 0.0001, "Learning rate used for training")
flags.DEFINE_float("weight_decay", 0.0001, "L2 regularization strength.")
flags.DEFINE_enum(
    "nn_api_version",
    "linen",
    ["linen", "nnx"],
    "What type of flax api should be used for training?.                 "
    " Currently, linen and nnx are supported",
)

flags.DEFINE_bool("verbose", False, "Print information about the model.")
flags.DEFINE_bool(
    "visualise", False, "Use it to draw a treescope layout of the model"
)


flags.mark_flag_as_required("game")
flags.mark_flag_as_required("path")


def build_model() -> nnx.Module | linen.Module:
  """Builds a model."""
  game = pyspiel.load_game(FLAGS.game)

  config = {}
  if FLAGS.config_path is not None and os.path.exists(FLAGS.config_path):
    with open(FLAGS.config_path, "r") as f:
      config = json.load(f)

  model = utils.api_selector(
      config.get("nn_api_version", FLAGS.nn_api_version)
  ).Model.build_model(
      config.get("nn_model", FLAGS.nn_model),
      game.observation_tensor_shape(),
      game.num_distinct_actions(),
      config.get("nn_width", FLAGS.nn_width),
      config.get("nn_depth", FLAGS.nn_depth),
      config.get("weight_decay", FLAGS.weight_decay),
      config.get("learning_rate", FLAGS.learning_rate),
      config.get("path", FLAGS.path),
  )

  if FLAGS.checkpoint_step is not None:
    model.load_checkpoint(FLAGS.checkpoint_step)

  return model


def main(unused_argv):
  model = build_model()

  if FLAGS.visualise:
    nnx_model = (
        utils.linen_to_nnx(model._model)  # pylint: disable=protected-access
        if not isinstance(model, nnx.Module)
        else model
    )
    nnx.display(nnx_model)
  else:
    # In essence, just prepared the model graph for training
    model.save_checkpoint(0)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A basic AlphaZero implementation.

This implements the AlphaZero training algorithm. It spawns N actors which feed
trajectories into a replay buffer which are consumed by a learner. The learner
generates new weights, saves a checkpoint, and tells the actors to update. There
are also M evaluators running games continuously against a standard MCTS+Solver,
though each at a different difficulty (i.e. number of simulations for MCTS).

Due to the multi-process nature of this algorithm the logs are written to files,
one per process. The learner logs are also output to stdout. The checkpoints are
also written to the same directory.

Links to relevant articles/papers:
  https://deepmind.com/blog/article/alphago-zero-starting-scratch has an open
    access link to the AlphaGo Zero nature paper.
  https://deepmind.com/blog/article/alphazero-shedding-new-light-grand-games-chess-shogi-and-go
    has an open access link to the AlphaZero science paper.
"""

import datetime
import functools
import itertools
import json
import multiprocessing
import os
import sys
import tempfile
import time
import traceback
from typing import Any, Callable

import chex
import jax.numpy as jnp
import numpy as np

from open_spiel.python.algorithms import mcts
from open_spiel.python.algorithms.alpha_zero import evaluator as evaluator_lib
from open_spiel.python.algorithms.alpha_zero import replay_buffer as buffer_lib
from open_spiel.python.algorithms.alpha_zero import utils
import pyspiel
from open_spiel.python.utils import data_logger
from open_spiel.python.utils import file_logger
from open_spiel.python.utils import spawn
from open_spiel.python.utils import stats

# pylint: disable=g-bare-generic

# Set the start method to 'spawn' for CUDA compatibility and monkey-patch
# to prevent OpenSpiel from overriding it.
# See:
# https://github.com/google-deepmind/open_spiel/issues/1326
try:
  # jax and TF aren't compatible with the `fork` strategy due to the inner
  # multitheading
  # see:
  # https://github.com/jax-ml/jax/issues/1805
  multiprocessing.set_start_method("spawn", force=True)

  # Monkey-patch to prevent OpenSpiel from changing the start method.
  def _do_nothing_set_start_method(method, force=False):
    del method, force
    pass

  multiprocessing.set_start_method = _do_nothing_set_start_method
except RuntimeError:
  # This may be raised in child processes where the context is already set.
  pass

# Time to wait for processes to join.
JOIN_WAIT_DELAY = 0.001


@chex.dataclass(frozen=True)
class TrajectoryState:
  """A particular point along a trajectory."""

  observation: chex.Array
  current_player: chex.Array
  legals_mask: chex.Array
  action: chex.Array
  policy: chex.Array
  value: chex.Array


@chex.dataclass(frozen=True)
class Trajectory:
  states: list[TrajectoryState]
  returns: chex.Array


@chex.dataclass(frozen=True)
class Config:
  """A config for the model/experiment."""

  game: str
  path: str
  learning_rate: float
  weight_decay: float
  decouple_weight_decay: bool
  train_batch_size: int
  replay_buffer_size: int
  replay_buffer_reuse: bool
  max_steps: int
  checkpoint_freq: int
  actors: int

  evaluators: int
  evaluation_window: int
  eval_levels: int
  uct_c: float
  max_simulations: int

  policy_alpha: float
  policy_epsilon: float
  temperature: float
  temperature_drop: float

  nn_model: str
  nn_width: int
  nn_depth: int
  observation_shape: chex.Shape
  output_size: int
  verbose: bool
  quiet: bool

  nn_api_version: str = "nnx"


def _init_model_from_config(config: Config):
  return utils.api_selector(config.nn_api_version).Model.build_model(
      config.nn_model,
      config.observation_shape,
      config.output_size,
      config.nn_width,
      config.nn_depth,
      config.weight_decay,
      config.learning_rate,
      config.path,
      decouple_weight_decay=config.decouple_weight_decay,
  )


def watcher(fn):
  """A decorator to fn/processes that gives a logger and logs exceptions."""

  @functools.wraps(fn)
  def _watcher(*, config, num=None, **kwargs):
    """Wrap the decorated function."""
    name = fn.__name__
    if num is not None:
      name += "-" + str(num)
    with file_logger.FileLogger(config.path, name, config.quiet) as logger:
      print("{} started".format(name))
      logger.print("{} started".format(name))
      try:
        return fn(config=config, logger=logger, **kwargs)
      except Exception as e:
        logger.print(
            "\n".join([
                "",
                " Exception caught ".center(60, "="),
                traceback.format_exc(),
                "=" * 60,
            ])
        )
        print("Exception caught in {}: {}".format(name, e))
        raise
      finally:
        logger.print("{} exiting".format(name))
        print("{} exiting".format(name))

  return _watcher


def _init_bot(
    config: Config, game: Any, evaluator_: mcts.Evaluator, evaluation: bool
) -> mcts.MCTSBot:
  """Initialises a MCTS bot."""
  noise = None if evaluation else (config.policy_epsilon, config.policy_alpha)
  return mcts.MCTSBot(
      game,
      config.uct_c,
      config.max_simulations,
      evaluator_,
      solve=False,
      dirichlet_noise=noise,
      child_selection_fn=mcts.SearchNode.puct_value,
      verbose=config.verbose,
      dont_return_chance_node=True,
  )


def _play_game(
    logger: Any,
    game_num: int,
    game: Any,
    bots: tuple[mcts.MCTSBot, mcts.MCTSBot],
    temperature: float,
    temperature_drop: int,
) -> Trajectory:
  """Play one game, return the trajectory."""
  trajectory_states = []
  actions = []
  state = game.new_initial_state()
  random_state = np.random.RandomState()
  logger.opt_print(" Starting game {} ".format(game_num).center(60, "-"))
  logger.opt_print("Initial state:\n{}".format(state))
  while not state.is_terminal():
    if state.is_chance_node():
      # For chance nodes, rollout according to chance node's probability
      # distribution
      outcomes = state.chance_outcomes()
      action_list, prob_list = zip(*outcomes)
      action = random_state.choice(action_list, p=prob_list)
      action_str = state.action_to_string(state.current_player(), action)
      actions.append(action_str)
      state.apply_action(action)
    else:
      root = bots[state.current_player()].mcts_search(state)
      policy = np.zeros(game.num_distinct_actions())
      for c in root.children:
        policy[c.action] = c.explore_count
      policy = policy ** (1 / temperature)
      policy /= policy.sum()
      if len(actions) >= temperature_drop:
        action = root.best_child().action
      else:
        action = np.random.choice(len(policy), p=policy)

      trajectory_states.append(
          TrajectoryState(
              observation=state.observation_tensor(),
              current_player=state.current_player(),
              legals_mask=state.legal_actions_mask(),
              action=action,
              policy=policy,
              value=root.total_reward / root.explore_count,
          )
      )
      action_str = state.action_to_string(state.current_player(), action)
      actions.append(action_str)
      logger.opt_print(
          f"Player {state.current_player()} sampled action: {action_str}"
      )
      state.apply_action(action)
  logger.opt_print("Next state:\n{}".format(state))

  trajectory = Trajectory(states=trajectory_states, returns=state.returns())

  logger.print(
      "Game {}: Returns: {}; Actions: {}".format(
          game_num, " ".join(map(str, trajectory.returns)), " ".join(actions)
      )
  )
  return trajectory


def update_checkpoint(
    logger: Any,
    queue: spawn._ProcessQueue,
    model: Any,
    az_evaluator: evaluator_lib.AlphaZeroEvaluator,
) -> bool:
  """Read the queue for a checkpoint to load, or an exit signal."""
  path = None
  while True:  # Get the last message, ignore intermediate ones.
    try:
      path = queue.get_nowait()
    except spawn.Empty:
      break
  if path:
    logger.print("Inference cache:", az_evaluator.cache_info())
    logger.print("Loading checkpoint", path)
    model.load_checkpoint(path)
    az_evaluator.clear_cache()
  elif path is not None:  # Empty string means stop this process.
    return False
  return True


@watcher
def actor(
    *,
    config: Config,
    game: pyspiel.Game,
    logger: Any,
    queue: spawn._ProcessQueue,
) -> None:
  """An actor process runner that generates games and returns trajectories."""
  logger.print("Initializing model")
  model = _init_model_from_config(config)
  logger.print("Initializing bots")
  az_evaluator = evaluator_lib.AlphaZeroEvaluator(game, model)
  bots = [
      _init_bot(config, game, az_evaluator, False),
      _init_bot(config, game, az_evaluator, False),
  ]
  for game_num in itertools.count(1):
    if not update_checkpoint(logger, queue, model, az_evaluator):
      return
    queue.put(
        _play_game(
            logger,
            game_num,
            game,
            bots,
            config.temperature,
            config.temperature_drop,
        )
    )


@watcher
def evaluator(
    *, game: pyspiel.Game, config: Config, logger: Any, queue
) -> None:
  """A process that plays the latest checkpoint vs standard MCTS."""
  results = buffer_lib.Buffer(config.evaluation_window, force_cpu=True)

  logger.print("Initializing model")
  model = _init_model_from_config(config)
  logger.print("Initializing bots")
  az_evaluator = evaluator_lib.AlphaZeroEvaluator(game, model)
  random_evaluator = mcts.RandomRolloutEvaluator()

  for game_num in itertools.count():
    if not update_checkpoint(logger, queue, model, az_evaluator):
      return

    az_player = game_num % 2
    difficulty = (game_num // 2) % config.eval_levels
    max_simulations = int(config.max_simulations * (10 ** (difficulty / 2)))
    bots = [
        _init_bot(config, game, az_evaluator, True),
        mcts.MCTSBot(
            game,
            config.uct_c,
            max_simulations,
            random_evaluator,
            solve=True,
            verbose=config.verbose,
            dont_return_chance_node=True,
        ),
    ]
    if az_player == 1:
      bots = list(reversed(bots))

    trajectory = _play_game(
        logger, game_num, game, bots, temperature=1, temperature_drop=0
    )
    results.append(jnp.asarray(trajectory.returns[az_player]))
    queue.put((difficulty, jnp.asarray(trajectory.returns[az_player])))

    logger.print(
        f"AZ: {trajectory.returns[az_player]},      MCTS:"
        f" {trajectory.returns[1 - az_player]},      AZ avg/{len(results)}:"
        f" {jnp.mean(results.data):.3f}"
    )


@watcher
def learner(
    *,
    game: pyspiel.Game,
    config: Config,
    actors: list[spawn.Process],
    evaluators: list[spawn.Process],
    broadcast_fn: Callable,
    logger: Any,
) -> None:
  """A learner that consumes the replay buffer and trains the network."""
  logger.also_to_stdout = True

  replay_buffer = buffer_lib.Buffer(config.replay_buffer_size)
  learn_rate = config.replay_buffer_size // config.replay_buffer_reuse
  logger.print("Initializing model")
  model = _init_model_from_config(config)
  logger.print(
      f"Model type: {config.nn_model}({config.nn_width}, {config.nn_depth})"
  )
  logger.print("Model size:", model.num_trainable_variables, "variables")

  save_path = model.save_checkpoint(0)
  logger.print("Initial checkpoint:", save_path)
  broadcast_fn(str(save_path))

  data_log = data_logger.DataLoggerJsonLines(config.path, "learner", True)

  stage_count = config.eval_levels
  value_accuracies = [stats.BasicStats() for _ in range(stage_count)]
  value_predictions = [stats.BasicStats() for _ in range(stage_count)]
  game_lengths = stats.BasicStats()
  game_lengths_hist = stats.HistogramNumbered(game.max_game_length() + 1)
  outcomes = stats.HistogramNamed(["Player1", "Player2", "Draw"])
  # evaluation is yet on cpu
  evals = [
      buffer_lib.Buffer(config.evaluation_window, force_cpu=True)
      for _ in range(config.eval_levels)
  ]
  total_trajectories = 0

  def trajectory_generator():
    """Merge all the actor queues into a single generator."""
    while True:
      found = 0
      for actor_process in actors:
        try:
          yield actor_process.queue.get_nowait()
        except spawn.Empty:
          pass
        else:
          found += 1
      if found == 0:
        time.sleep(0.01)  # 10ms

  def collect_trajectories() -> tuple[int, int]:
    """Collects the trajectories from actors into the replay buffer."""

    logger.print("Collecting trajectories")
    num_trajectories = 0
    num_states = 0
    for trajectory in trajectory_generator():
      num_trajectories += 1
      num_states += len(trajectory.states)
      game_lengths.add(len(trajectory.states))
      game_lengths_hist.add(len(trajectory.states))

      # we learn from perspective of only the first player,
      # rather than rotating
      game_outcome = trajectory.returns[0]
      if game_outcome > 0:
        outcomes.add(0)
      elif game_outcome < 0:
        outcomes.add(1)
      else:
        outcomes.add(2)

      for s in trajectory.states:
        replay_buffer.append(
            utils.TrainInput(
                observation=jnp.asarray(s.observation, dtype=jnp.float32),
                legals_mask=jnp.asarray(s.legals_mask, dtype=jnp.bool),
                policy=jnp.asarray(s.policy, dtype=jnp.float32),
                value=jnp.asarray(game_outcome, dtype=jnp.float32),
            )
        )

      for stage in range(stage_count):
        # Scale for the length of the game
        index = (len(trajectory.states) - 1) * stage // (stage_count - 1)
        n = trajectory.states[index]
        # Replaced by the MSE to get a smoother estimate of the value learning
        value_accuracies[stage].add(
            (n.value - trajectory.returns[n.current_player]) ** 2
        )
        value_predictions[stage].add(abs(n.value))

      if num_states >= learn_rate:
        break
    return num_trajectories, num_states

  def learn(step: int | str) -> tuple[str, utils.Losses]:
    """Sample from the replay buffer, update weights and save a checkpoint."""
    losses = []
    for _ in range(len(replay_buffer) // config.train_batch_size):
      data = replay_buffer.sample(config.train_batch_size)
      losses.append(model.update(data))

    # Always save a checkpoint, either for keeping or for loading the weights to
    # the actors. We only allow numbers, so use -1 as the "latest".
    save_path = model.save_checkpoint(
        step if step % config.checkpoint_freq == 0 else -1
    )

    # for an unlucky case when the agent hasn't collected enough transitions
    if losses:
      losses = sum(losses, utils.Losses(policy=0, value=0, l2=0)) / len(losses)
    else:
      losses = utils.Losses(policy=0, value=0, l2=0)

    logger.print(losses)
    logger.print("Checkpoint saved:", save_path)
    return save_path, losses

  last_time = time.time() - 60
  for step in itertools.count(1):
    for value_accuracy in value_accuracies:
      value_accuracy.reset()
    for value_prediction in value_predictions:
      value_prediction.reset()
    game_lengths.reset()
    game_lengths_hist.reset()
    outcomes.reset()

    num_trajectories, num_states = collect_trajectories()
    total_trajectories += num_trajectories
    now = time.time()
    seconds = now - last_time
    last_time = now

    logger.print("Step:", step)
    logger.print((
        f"Collected {num_states:5} states from {num_trajectories:3} games, "
        f"{(num_states / seconds):.1f} states/s. "
        f"{(num_states / (config.actors * seconds)):.1f} states/(s*actor), "
        f"game length: {(num_states / num_trajectories):.1f}"
    ))
    logger.print(
        f"Buffer size: {len(replay_buffer)}. States seen:"
        f" {replay_buffer.total_seen}"
    )

    save_path, losses = learn(step)

    for eval_process in evaluators:
      while True:
        try:
          difficulty, outcome = eval_process.queue.get_nowait()
          evals[difficulty].append(outcome)
        except spawn.Empty:
          break

    batch_size_stats = stats.BasicStats()  # Only makes sense in C++.
    batch_size_stats.add(1)
    data_log.write({
        "step": step,
        "total_states": replay_buffer.total_seen,
        "states_per_s_actor": num_states / (config.actors * seconds),
        "total_trajectories": total_trajectories,
        "trajectories_per_s": num_trajectories / seconds,
        "queue_size": 0,  # Only available in C++.
        "game_length": game_lengths.as_dict,
        "game_length_hist": game_lengths_hist.data,
        "outcomes": outcomes.data,
        "value_accuracy": [v.as_dict for v in value_accuracies],
        "value_prediction": [v.as_dict for v in value_predictions],
        "eval": {
            "count": evals[0].total_seen,
            "results": [np.mean(e.data).item() if e else 0 for e in evals],
        },
        "batch_size": batch_size_stats.as_dict,
        "batch_size_hist": [0, 1],
        "loss": {
            "policy": float(losses.policy),
            "value": float(losses.value),
            "l2reg": float(losses.l2),
            "sum": float(losses.total),
        },
        "cache": {  # Null stats because it's hard to report between processes.
            "size": 0,
            "max_size": 0,
            "usage": 0,
            "requests": 0,
            "requests_per_s": 0,
            "hits": 0,
            "misses": 0,
            "misses_per_s": 0,
            "hit_rate": 0,
        },
    })
    logger.print()

    if config.max_steps > 0 and step >= config.max_steps:
      break

    broadcast_fn(str(save_path))


def alpha_zero(config: Config) -> None:
  # NOTE: a single device accelearation is currently supported

  """Start all the worker processes for a full alphazero setup."""
  game = pyspiel.load_game(config.game)
  config = config.replace(
      observation_shape=game.observation_tensor_shape(),
      output_size=game.num_distinct_actions(),
  )

  print("Starting game", config.game)
  if game.num_players() != 2:
    sys.exit("AlphaZero can only handle 2-player games.")
  game_type = game.get_type()
  if game_type.reward_model != pyspiel.GameType.RewardModel.TERMINAL:
    raise ValueError("Game must have terminal rewards.")
  if game_type.dynamics != pyspiel.GameType.Dynamics.SEQUENTIAL:
    raise ValueError("Game must have sequential turns.")

  path = config.path
  if not path:
    path = tempfile.mkdtemp(
        prefix="az-{}-{}-".format(
            datetime.datetime.now().strftime("%Y-%m-%d-%H-%M"), config.game
        )
    )
    config = config.replace(path=path)

  if not os.path.exists(path):
    os.makedirs(path)
  if not os.path.isdir(path):
    sys.exit(f"{path} isn't a directory")
  print(f"Writing logs and checkpoints to: {path}")
  print(
      f"Model type: {config.nn_model}(width={config.nn_width},"
      f" depth={config.nn_depth})"
  )

  with open(os.path.join(config.path, "config.json"), "w") as fp:
    fp.write(json.dumps(config.__dict__, indent=2, sort_keys=True) + "\n")

  actors = [
      spawn.Process(actor, kwargs={"game": game, "config": config, "num": i})
      for i in range(config.actors)
  ]

  evaluators = [
      spawn.Process(
          evaluator, kwargs={"game": game, "config": config, "num": i}
      )
      for i in range(config.evaluators)
  ]

  def broadcast(msg):
    for proc in actors + evaluators:
      proc.queue.put(msg)

  try:
    learner(
        game=game,
        config=config,
        actors=actors,  # pylint: disable=missing-kwoa
        evaluators=evaluators,
        broadcast_fn=broadcast,
    )
  except (KeyboardInterrupt, EOFError):
    print("Caught a KeyboardInterrupt, stopping early.")
  finally:
    broadcast("")
    # for actor processes to join we have to make sure that their q_in is empty,
    # including backed up items
    for proc in actors:
      while proc.exitcode is None:
        while not proc.queue.empty():
          proc.queue.get_nowait()
        proc.join(JOIN_WAIT_DELAY)
    for proc in evaluators:
      proc.join()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for OpenSpiel 'Repeated' PokerkitWrapper."""

import itertools
import json
import logging
import random

from absl.testing import absltest
from absl.testing import parameterized

# Hack to handle an inability to import pokerkit gracefully. Necessary since
# OpenSpiel may support earlier versions of python that pokerkit doesn't.
IMPORTED_ALL_LIBRARIES = False
try:
  # pylint: disable=g-import-not-at-top
  import pokerkit

  from open_spiel.python.games import pokerkit_wrapper
  from open_spiel.python.games import repeated_pokerkit
  import pyspiel
  # pylint: disable=g-import-not-at-top
  IMPORTED_ALL_LIBRARIES = True
except ImportError as e:
  logging.error("Failed to import needed pokerkit libraries: %s", e)
  pokerkit_wrapper = None
  repeated_pokerkit = None
  pokerkit = None
  pyspiel = None

if IMPORTED_ALL_LIBRARIES:
  assert pokerkit is not None
  assert pokerkit_wrapper is not None
  assert repeated_pokerkit is not None
  assert pyspiel is not None

  ACTION_CHECK_OR_CALL = pokerkit_wrapper.ACTION_CHECK_OR_CALL
  ACTION_FOLD = pokerkit_wrapper.ACTION_FOLD

  Card = pokerkit.Card
  ACE = pokerkit.Rank.ACE
  DEUCE = pokerkit.Rank.DEUCE
  KING = pokerkit.Rank.KING
  QUEEN = pokerkit.Rank.QUEEN
  JACK = pokerkit.Rank.JACK
  TEN = pokerkit.Rank.TEN
  NINE = pokerkit.Rank.NINE
  EIGHT = pokerkit.Rank.EIGHT
  SEVEN = pokerkit.Rank.SEVEN
  SIX = pokerkit.Rank.SIX
  FIVE = pokerkit.Rank.FIVE
  FOUR = pokerkit.Rank.FOUR
  TREY = pokerkit.Rank.TREY
  SPADE = pokerkit.Suit.SPADE
  CLUB = pokerkit.Suit.CLUB
  HEART = pokerkit.Suit.HEART
  DIAMOND = pokerkit.Suit.DIAMOND

  # E.g.
  # [
  #   {variant: "NoLimitTexasHoldem"}
  #   {variant: "FixedLimitTexasHoldem"}
  #   ...
  # ]
  _ALL_VARIANTS = [
      {"variant": variant}
      for variant in pokerkit_wrapper.SUPPORTED_VARIANT_MAP.keys()
  ]
  # E.g.
  # [
  #   {variant: "NoLimitTexasHoldem", reset_stacks: True},
  #   {variant: "NoLimitTexasHoldem", reset_stacks: False},
  #   {variant: "FixedLimitTexasHoldem", reset_stacks: True},
  #   {variant: "FixedLimitTexasHoldem", reset_stacks: False},
  #   ...
  # ]
  _ALL_VARIANTS_CASH_AND_TOURNAMENT_PARAMS = [
      params | {"reset_stacks": reset_stacks}
      for params, reset_stacks in itertools.product(
          _ALL_VARIANTS, [True, False]
      )
  ]

  class RepeatedPokerkitTest(parameterized.TestCase):
    """Test the OpenSpiel game 'Repeated Pokerkit'."""

    def _load_repeated_pokerkit_game_with_variant(self, variant, params):
      """Loads a repeated_pokerkit game with variant-specific param filtering."""
      use_bring_in = variant in pokerkit_wrapper.VARIANT_PARAM_USAGE["bring_in"]
      use_blinds = (
          variant
          in pokerkit_wrapper.VARIANT_PARAM_USAGE["raw_blinds_or_straddles"]
      )
      use_fixed_bet_sizes = (
          variant in pokerkit_wrapper.VARIANT_PARAM_USAGE["small_bet"]
      )
      if use_fixed_bet_sizes:
        self.assertIn(variant, pokerkit_wrapper.VARIANT_PARAM_USAGE["big_bet"])

      if not use_bring_in:
        params["pokerkit_game_params"].pop("bring_in")
      if not use_blinds:
        self.assertTrue(use_bring_in and use_fixed_bet_sizes)
        params["pokerkit_game_params"].pop("blinds")
        params["rotate_dealer"] = False
      if not use_fixed_bet_sizes:
        self.assertTrue(use_blinds and not use_bring_in)
        params["pokerkit_game_params"].pop("small_bet")
        params["pokerkit_game_params"].pop("big_bet")

      return pyspiel.load_game("python_repeated_pokerkit", params)

    def test_parse_blind_schedule(self):
      blind_schedule_str = "1:1/2;2:2/3"
      blind_schedule = repeated_pokerkit.parse_blind_schedule(
          blind_schedule_str
      )
      self.assertLen(blind_schedule, 2)
      self.assertEqual(blind_schedule[0].num_hands, 1)
      self.assertEqual(blind_schedule[0].small_blind, 1)
      self.assertEqual(blind_schedule[0].big_blind, 2)
      self.assertEqual(blind_schedule[1].num_hands, 2)
      self.assertEqual(blind_schedule[1].small_blind, 2)
      self.assertEqual(blind_schedule[1].big_blind, 3)

    def test_parse_bet_size_schedule(self):
      bet_size_schedule_str = "1:10/20;2:20/30"
      bet_size_schedule = repeated_pokerkit.parse_bet_size_schedule(
          bet_size_schedule_str
      )
      self.assertLen(bet_size_schedule, 2)
      self.assertEqual(bet_size_schedule[0].num_hands, 1)
      self.assertEqual(bet_size_schedule[0].small_bet_size, 10)
      self.assertEqual(bet_size_schedule[0].big_bet_size, 20)
      self.assertEqual(bet_size_schedule[1].num_hands, 2)
      self.assertEqual(bet_size_schedule[1].small_bet_size, 20)
      self.assertEqual(bet_size_schedule[1].big_bet_size, 30)

    def test_parse_bring_in_schedule(self):
      bring_in_schedule_str = "1:5;2:10"
      bring_in_schedule = repeated_pokerkit.parse_bring_in_schedule(
          bring_in_schedule_str
      )
      self.assertLen(bring_in_schedule, 2)
      self.assertEqual(bring_in_schedule[0].num_hands, 1)
      self.assertEqual(bring_in_schedule[0].bring_in, 5)
      self.assertEqual(bring_in_schedule[1].num_hands, 2)
      self.assertEqual(bring_in_schedule[1].bring_in, 10)

    def test_random_sim_cash_game_headsup(self):
      # Note the reset_stacks=True
      game = pyspiel.load_game(
          "python_repeated_pokerkit(max_num_hands=20,reset_stacks=True,"
          "rotate_dealer=True,pokerkit_game_params=python_pokerkit_wrapper())"
      )
      pyspiel.random_sim_test(game, num_sims=5, serialize=False, verbose=True)

    def test_random_sim_tournament_headsup(self):
      # Note the reset_stacks=False
      game = pyspiel.load_game(
          "python_repeated_pokerkit(max_num_hands=20,reset_stacks=False,"
          "rotate_dealer=True,pokerkit_game_params=python_pokerkit_wrapper())"
      )
      pyspiel.random_sim_test(game, num_sims=5, serialize=False, verbose=True)

    @parameterized.parameters(*_ALL_VARIANTS_CASH_AND_TOURNAMENT_PARAMS)
    def test_random_sim_all_variants_cash_and_tournament_headsup(
        self,
        variant,
        # True -> cash game, False -> tournament
        reset_stacks,
    ):
      params = {
          "max_num_hands": 10,
          "reset_stacks": reset_stacks,
          "rotate_dealer": True,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "variant": variant,
              "num_players": 2,
              "blinds": "2 4",
              "bring_in": 4,
              "small_bet": 8,
              "big_bet": 16,
              "stack_sizes": "64 64",
          },
      }
      game = self._load_repeated_pokerkit_game_with_variant(variant, params)
      # NOTE: Consider carefully before increasing num_sims any further. As of
      # 2025, num_sims=3 is already pushing it (and e.g. num_sims=5 takes ~50%
      # longer than num_sims=3 does).
      pyspiel.random_sim_test(game, num_sims=3, serialize=False, verbose=False)

    @parameterized.parameters(*_ALL_VARIANTS_CASH_AND_TOURNAMENT_PARAMS)
    def test_random_sim_all_variants_cash_and_tournament_6max(
        self,
        variant,
        # True -> cash game, False -> tournament
        reset_stacks,
    ):
      params = {
          "max_num_hands": 3,
          "reset_stacks": reset_stacks,
          "rotate_dealer": True,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "variant": variant,
              "num_players": 6,
              "blinds": "2 4",
              "bring_in": 4,
              "small_bet": 8,
              "big_bet": 16,
              "stack_sizes": "30 30 30 30 30 30",
          },
      }
      game = self._load_repeated_pokerkit_game_with_variant(variant, params)
      # NOTE: Consider carefully before increasing num_sims any further. As of
      # 2025, num_sims=2 is already pushing it (and e.g. num_sims=3 takes ~50%
      # longer than num_sims=2 does).
      pyspiel.random_sim_test(game, num_sims=2, serialize=False, verbose=False)

    def test_blind_schedule_progression(self):
      blind_schedule_str = "1:10/20;2:20/40;1:50/100;1:1000/1000"
      params = {
          "max_num_hands": 100,
          "reset_stacks": False,
          "rotate_dealer": True,
          "blind_schedule": blind_schedule_str,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "num_players": 2,
              "blinds": "1 2",  # Initial blinds should be overridden
              "stack_sizes": "1000 1000",
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()

      # TODO: b/444333187 - consider refactoring to be "DAMP" instead of "DRY".
      expected_blinds = [
          (10, 20),  # Hand 0
          (20, 40),  # Hand 1
          (20, 40),  # Hand 2
          (50, 100),  # Hand 3
          (1000, 1000),  # Hands 4+
      ]
      seen_hands = set()
      current_hand = state._hand_number
      self.assertEqual(state._hand_number, 0)
      while not state.is_terminal():
        self.assertEqual(state._hand_number, current_hand)
        if current_hand not in seen_hands:
          seen_hands.add(current_hand)
          small_blind, big_blind = (
              expected_blinds[-1]
              if current_hand >= len(expected_blinds)
              else expected_blinds[current_hand]
          )
          self.assertEqual(
              state._small_blind,
              small_blind,
          )
          self.assertEqual(
              state._big_blind,
              big_blind,
          )

        if state.is_chance_node():
          action = random.choice([o for o, _ in state.chance_outcomes()])
        else:
          legal_actions = state.legal_actions()
          # Choose to check/call in all situations to ensure we reach the final
          # blind schedule level (at which point everyone will be forced
          # all-in).
          if ACTION_CHECK_OR_CALL in legal_actions:
            action = ACTION_CHECK_OR_CALL
          else:
            action = random.choice(state.legal_actions())

        state.apply_action(action)
        # apply_action handles the transition to the next hand when the
        # pokerkit_wrapper state is terminal.
        self.assertBetween(state._hand_number, current_hand, current_hand + 1)
        current_hand = state._hand_number

    def test_initially_scheduled_blinds_larger_than_stacks(self):
      blind_schedule_str = "1:200/400"
      params = {
          "max_num_hands": 1,
          "reset_stacks": False,
          "rotate_dealer": True,
          "blind_schedule": blind_schedule_str,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "num_players": 2,
              "blinds": "1 2",  # Initial blinds should be overridden
              "stack_sizes": "100 100",
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()
      self.assertEqual(state._hand_number, 0)
      self.assertEqual(state._small_blind, 200)
      self.assertEqual(state._big_blind, 400)
      self.assertEqual(state._stacks, [100, 100])
      # With 2 players, player 1 is dealer/SB and player 0 is BB in hand 0.
      # P1 posts SB=min(100, 200)=100, P0 posts BB=min(100, 400)=100.
      # Both players are all-in.
      # Pokerkit state stacks are seat-based. If P0=seat0=BB, P1=seat1=SB:
      # Seat 0 stack: 100-100=0. Seat 1 stack: 100-100=0.
      self.assertEqual(
          state.pokerkit_wrapper_state._wrapped_state.stacks, [0, 0]
      )
      self.assertEqual(
          state.pokerkit_wrapper_state._wrapped_state.bets, [100, 100]
      )

      # 4 chance nodes for hole cards
      for _ in range(4):
        self.assertTrue(state.is_chance_node())
        action = random.choice([o for o, _ in state.chance_outcomes()])
        state.apply_action(action)

      # After hole cards, betting round should be skipped as players are all in,
      # and board cards should be dealt.
      # 5 chance nodes for board cards
      for _ in range(5):
        self.assertTrue(state.is_chance_node())
        action = random.choice([o for o, _ in state.chance_outcomes()])
        state.apply_action(action)

      self.assertTrue(state.is_terminal())

    def test_later_stage_blinds_larger_than_stacks(self):
      blind_schedule_str = "1:1/2;1:200/400"
      params = {
          "max_num_hands": 5,
          "reset_stacks": False,
          "rotate_dealer": True,
          "blind_schedule": blind_schedule_str,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "num_players": 2,
              "blinds": "1 2",
              "stack_sizes": "100 100",
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()
      state_struct = state.to_struct()
      self.assertEqual(state_struct.hand_number, 0)
      self.assertEqual(state_struct.stacks, [100, 100])
      self.assertEqual(state_struct.dealer, 1)
      self.assertEqual(state_struct.small_blind, 1)
      self.assertEqual(state_struct.big_blind, 2)

      # 4 chance nodes for hole cards
      for _ in range(4):
        self.assertTrue(state.is_chance_node())
        action = random.choice([o for o, _ in state.chance_outcomes()])
        state.apply_action(action)

      # SB immediately folds preflop to start the second hand with higher
      # blinds.
      state.apply_action(ACTION_FOLD)

      self.assertFalse(state.is_terminal())
      self.assertTrue(state.is_chance_node())
      state_struct = state.to_struct()
      self.assertEqual(state_struct.hand_number, 1)
      self.assertEqual(state_struct.stacks, [101, 99])
      self.assertEqual(state_struct.dealer, 0)
      self.assertEqual(state_struct.small_blind, 200)
      self.assertEqual(state_struct.big_blind, 400)

      # 4 chance nodes for hole cards first, but then afterwards betting round
      # should be skipped as players are all in, and board cards should be
      # dealt, meaning +5 more chance nodes => 9 total chance nodes.
      #
      # Unlike before, we can't have a player fold so to have a deterministic
      # winner we need to choose the specific cards ourselves.
      # Result should be AA778 winning vs 77668 (with the 22 being counterfeit)
      dealt_cards = [
          Card(ACE, SPADE),
          Card(DEUCE, SPADE),
          Card(ACE, HEART),
          Card(DEUCE, DIAMOND),
          Card(SIX, CLUB),  # Flop 1
          Card(SIX, HEART),  # Flop 2
          Card(SEVEN, SPADE),  # Flop 3
          Card(SEVEN, HEART),  # Turn
          Card(EIGHT, DIAMOND),  # River
      ]
      for card in dealt_cards:
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[card])

      self.assertFalse(state.is_terminal())
      state_struct = state.to_struct()
      self.assertEqual(state_struct.hand_number, 2)
      self.assertEqual(state_struct.stacks, [2, 198])
      self.assertEqual(state_struct.small_blind, 200)
      self.assertEqual(state_struct.big_blind, 400)

      # Same again as before, but to accomodate dealer button rotation we have
      # to reverse the hole cards.
      dealt_cards = [
          Card(DEUCE, SPADE),
          Card(ACE, SPADE),
          Card(DEUCE, DIAMOND),
          Card(ACE, HEART),
          Card(SIX, CLUB),  # Flop 1
          Card(SIX, HEART),  # Flop 2
          Card(SEVEN, SPADE),  # Flop 3
          Card(SEVEN, HEART),  # Turn
          Card(EIGHT, DIAMOND),  # River
      ]
      for card in dealt_cards:
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[card])

      self.assertTrue(state.is_terminal())
      state_struct = state.to_struct()
      self.assertEqual(state_struct.hand_number, 3)
      self.assertEqual(state_struct.stacks, [0, 200])
      self.assertEqual(state_struct.small_blind, 200)
      self.assertEqual(state_struct.big_blind, 400)

    def test_stacks_carried_over(self):
      params = {
          "max_num_hands": 3,
          "reset_stacks": False,
          "rotate_dealer": True,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "num_players": 3,
              "blinds": "10 20",
              "stack_sizes": "100 100 100",  # Players 0 1 2: SB BB BTN
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()

      # Hand 0: P0 will win, P1 will bust, P2 will not be affected:

      self.assertEqual(state._hand_number, 0)
      self.assertEqual(state._stacks, [100, 100, 100])
      # Play a hand where P0 wins all of P1's chips and none of P2's chips.
      # P0 (SB) posts 10, P1 (BB) posts 20

      # First hole card (P0, then P1, then P2)
      state.apply_action(
          state.pokerkit_wrapper_game.card_to_int[Card(ACE, SPADE)]
      )
      state.apply_action(
          state.pokerkit_wrapper_game.card_to_int[Card(DEUCE, SPADE)]
      )
      state.apply_action(
          state.pokerkit_wrapper_game.card_to_int[Card(KING, HEART)]
      )

      # Second hole card (P0, then P1, then P2)
      # P0 has pocket Aces, P1 pockets Twos, P2 pocket Kings.
      state.apply_action(
          state.pokerkit_wrapper_game.card_to_int[Card(ACE, CLUB)]
      )
      state.apply_action(
          state.pokerkit_wrapper_game.card_to_int[Card(DEUCE, CLUB)]
      )
      state.apply_action(
          state.pokerkit_wrapper_game.card_to_int[Card(KING, SPADE)]
      )

      # P2 folds preflop
      state.apply_action(ACTION_FOLD)
      # P0 calls to 20 (10 additional), P1 checks
      state.apply_action(ACTION_CHECK_OR_CALL)
      state.apply_action(ACTION_CHECK_OR_CALL)
      # Flop
      state.apply_action(
          state.pokerkit_wrapper_game.card_to_int[Card(SEVEN, SPADE)]
      )
      state.apply_action(
          state.pokerkit_wrapper_game.card_to_int[Card(SEVEN, HEART)]
      )
      state.apply_action(
          state.pokerkit_wrapper_game.card_to_int[Card(SEVEN, DIAMOND)]
      )
      # P0 bets 80 (all-in), P1 calls
      state.apply_action(80)
      state.apply_action(ACTION_CHECK_OR_CALL)
      # Turn/River
      state.apply_action(
          state.pokerkit_wrapper_game.card_to_int[Card(SIX, CLUB)]
      )
      state.apply_action(
          state.pokerkit_wrapper_game.card_to_int[Card(SIX, DIAMOND)]
      )
      assert not state.is_terminal()

      # P0 wins the pot. P0 stack: 100 + 100 = 200. P1 stack: 100 - 100 = 0.
      # P2 stack is unchanged. It's now the next hand and only P0 and P2 are
      # still playing (ie in the underlying pokerkit game).
      self.assertEqual(state._hand_number, 1)
      self.assertEqual(state._stacks, [200, 0, 100])
      # Button rotates from P2 to P0 since P0 is still active. But, this is
      # heads up now, meaning:
      # - P0 (former SB) now SB/BTN: 200 - 10 => 190
      # - P2 (former BTN) now BB: 100 - 20 => 80
      self.assertEqual(
          state.pokerkit_wrapper_state._wrapped_state.stacks,
          # NOTE: This is the pokerkit's stacks, meaning that the
          # last element is the BTN (or in this case, SB/BTN). So
          # this makes sense given the rotation from the prior hand.
          [80, 190],
      )

    def test_stack_rotation_multiway(self):
      params = {
          "max_num_hands": 10,
          "reset_stacks": False,
          "rotate_dealer": True,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "variant": "NoLimitTexasHoldem",
              "num_players": 4,
              "blinds": "1 2",
              "stack_sizes": "200 200 200 200",
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()
      # Each time Seat 0 / SB will win the hand. So in the next hand we should
      # be seeing the BTN / Seat 3 (which is their new seat post rotation)
      # gaining the chips, while everyone else has lost some.
      board = [
          Card(ACE, DIAMOND),
          Card(KING, DIAMOND),
          Card(QUEEN, DIAMOND),
          Card(JACK, HEART),
          Card(SEVEN, HEART),
      ]
      hole_cards = [
          Card(ACE, SPADE),  # SB card 0
          Card(KING, SPADE),
          Card(QUEEN, SPADE),
          Card(JACK, SPADE),
          Card(ACE, CLUB),  # SB Card 1
          Card(KING, CLUB),
          Card(QUEEN, CLUB),
          Card(JACK, CLUB),
      ]
      expected_dealers = [3, 0, 1, 2, 3, 0, 1, 2, 3, 0]

      # RepeatedPokerkit only updates stacks at the end of each hand and keeps
      # them all unrotated.
      expected_stacks_repeated_pokerkit = [
          [200, 200, 200, 200],
          [215, 195, 195, 195],  # P0 won last hand (as SB), + 15 chips
          [210, 210, 190, 190],  # P1 won last hand (as SB), + 15 chips
          [205, 205, 205, 185],  # P2 won last hand (as SB), + 15 chips
          # 'Symmetry' of win/loss naturally results in us reaching the same
          # game state we had at the start.
          [200, 200, 200, 200],  # P3 won last hand (as SB), + 15 chips
          [215, 195, 195, 195],  # P0 won last hand (as SB), + 15 chips
          [210, 210, 190, 190],  # P1 won last hand (as SB), + 15 chips
          [205, 205, 205, 185],  # P2 won last hand (as SB), + 15 chips
          # Again, same symmetry.
          [200, 200, 200, 200],  # P3 won last hand (as SB), + 15 chips
          [215, 195, 195, 195],  # P0 won last hand (as SB), + 15 chips
          [210, 210, 190, 190],  # P1 won last hand (as SB), + 15 chips
      ]
      # The PokerkitWrapper however will have rotated, plus by the time we're
      # checking, already removed additional chips to handle the blinds.
      expected_stacks_pokerkit_wrapper = [
          [199, 198, 200, 200],
          [194, 193, 195, 215],
          [189, 188, 210, 210],
          [184, 203, 205, 205],
          # 'Symmetry' of win/loss naturally results in us reaching the same
          # game state we had at the start.
          [199, 198, 200, 200],
          [194, 193, 195, 215],
          [189, 188, 210, 210],
          [184, 203, 205, 205],
          # Again, same symmetry.
          [199, 198, 200, 200],
          [194, 193, 195, 215],
          # After the tenth hand we reach max_num_hands and so won't ever have
          # started the next hand, meaning the rotation will still match the
          # last hand and blinds won't have been removed. Hence this differs
          # from above.
          # (Note however how the *rotation* still differs from the unrotated
          # RepeatedPokerkit's stacks above at this point!)
          [210, 190, 190, 210],
      ]

      for i in range(10):
        self.assertEqual(state._hand_number, i)
        self.assertEqual(state._dealer, expected_dealers[i])

        # RepeatedPokerkit's stacks should only care about the results following
        # the end of the prior hand + generally are not rotated.
        self.assertEqual(state._stacks, expected_stacks_repeated_pokerkit[i])
        self.assertEqual(
            state.to_struct().stacks,
            expected_stacks_repeated_pokerkit[i],
        )
        # Pokerkit's stacks should include the exact current situation in the
        # current hand, meaning they will have already removed chips coming from
        # the blinds AND will also be rotated.
        self.assertEqual(
            state.pokerkit_wrapper_state._wrapped_state.stacks,
            expected_stacks_pokerkit_wrapper[i],
        )
        self.assertEqual(
            state.to_struct().pokerkit_state_struct.stacks,
            expected_stacks_pokerkit_wrapper[i],
        )

        # Deal hole cards: 4 players * 2 cards = 8 chance nodes
        for i in range(8):
          self.assertTrue(state.is_chance_node())
          state.apply_action(
              state.pokerkit_wrapper_game.card_to_int[hole_cards[i]]
          )

        # Preflop betting: call around to BB
        self.assertFalse(state.is_chance_node())
        for _ in range(4):
          state.apply_action(ACTION_CHECK_OR_CALL)

        # Flop
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[board[0]])
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[board[1]])
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[board[2]])

        # Flop betting: first player bets 3, everyone calls. This means all
        # players have contributed 5 chips total.
        state.apply_action(3)
        state.apply_action(ACTION_CHECK_OR_CALL)
        state.apply_action(ACTION_CHECK_OR_CALL)
        state.apply_action(ACTION_CHECK_OR_CALL)

        # Turn
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[board[3]])
        # Turn betting: check around
        for _ in range(4):
          state.apply_action(ACTION_CHECK_OR_CALL)

        # River
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[board[4]])
        # River betting: check around
        for _ in range(4):
          state.apply_action(ACTION_CHECK_OR_CALL)

      self.assertEqual(state._hand_number, 9)
      self.assertTrue(state.is_terminal())

      final_stacks_index = 10
      assert final_stacks_index == len(expected_stacks_repeated_pokerkit) - 1
      self.assertEqual(
          state._stacks, expected_stacks_repeated_pokerkit[final_stacks_index]
      )
      self.assertEqual(
          state.to_struct().stacks,
          expected_stacks_repeated_pokerkit[final_stacks_index],
      )
      self.assertEqual(
          state.pokerkit_wrapper_state._wrapped_state.stacks,
          expected_stacks_pokerkit_wrapper[final_stacks_index],
      )
      self.assertEqual(
          state.to_struct().pokerkit_state_struct.stacks,
          expected_stacks_pokerkit_wrapper[final_stacks_index],
      )

    def test_eliminates_players_at_exactly_zero_chips(self):
      params = {
          "max_num_hands": 3,
          "reset_stacks": False,
          "rotate_dealer": True,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "num_players": 3,
              "blinds": "10 20",
              "stack_sizes": "100 100 100",  # Players 0 1 2: SB BB BTN
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()

      # Deal each player two random cards (choice doesn't matter)
      # 3 players * 2 hole cards => 6 chance nodes
      for _ in range(6):
        action = random.choice([o for o, _ in state.chance_outcomes()])
        state.apply_action(action)

      # P3 BTN raises to *almost* all-in (1 chip behind)
      state.apply_action(99)
      # P1 SB Shoves all-in
      state.apply_action(100)
      # P2 BB folds => 80 chips left
      state.apply_action(ACTION_FOLD)
      # BTN folds => 1 chip left (should not bust). BB wins the pot and the next
      # hand starts.
      self.assertEqual(state._hand_number, 0)
      state.apply_action(ACTION_FOLD)
      self.assertEqual(state._hand_number, 1)
      self.assertEqual(state._stacks, [219, 80, 1])

      # Ensure *three* players here are still active, not just two.
      self.assertEqual(state.num_players(), 3)
      self.assertEqual(state._num_active_players, 3)
      # Similarly: 3 players * 2 hole cards => 6 chance nodes
      for _ in range(6):
        self.assertTrue(state.is_chance_node())
        self.assertEqual(state.current_player(), pyspiel.PlayerId.CHANCE)
        action = random.choice([o for o, _ in state.chance_outcomes()])
        state.apply_action(action)

        # Double check that nothing unexpectedly changes out from underneath
        # us here as we apply actions.
        self.assertEqual(state.num_players(), 3)
        self.assertEqual(state._num_active_players, 3)

      # P1 BTN bets 40
      state.apply_action(40)
      # P2 SB calls
      state.apply_action(ACTION_CHECK_OR_CALL)
      # P3 BB was already all-in due to having less than one Big Blind, so
      # gameplay should immediately procede to dealing the flop despite there
      # being 3 players.
      self.assertTrue(state.is_chance_node())

      # Definitely excessive, but triple checking this again just to be certain.
      self.assertEqual(state.num_players(), 3)
      self.assertEqual(state._num_active_players, 3)

    def test_multiway_all_but_one_players_bust_simultaneously(self):
      params = {
          "max_num_hands": 2,
          "reset_stacks": False,
          "rotate_dealer": True,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "num_players": 3,
              "blinds": "10 20",
              "stack_sizes": "100 100 100",  # Players 0 1 2: SB BB BTN
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()

      # Hand 0: P0 wins, P1 and P2 bust.
      # 3 players: P0=SB, P1=BB, P2=BTN
      self.assertEqual(state._hand_number, 0)
      self.assertEqual(state._stacks, [100, 100, 100])

      # Hole cards:
      # P0: Ac As
      # P1: Kc Ks
      # P2: Qc Qs
      hole_cards = [
          Card(ACE, CLUB),
          Card(KING, CLUB),
          Card(QUEEN, CLUB),
          Card(ACE, SPADE),
          Card(KING, SPADE),
          Card(QUEEN, SPADE),
      ]
      for card in hole_cards:
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[card])

      self.assertEqual(state.current_player(), 2)  # P2 BTN acts first preflop

      # P2 BTN raises all-in to 100
      state.apply_action(100)
      # P0 SB calls all-in (has 90 behind, action is for 100)
      self.assertEqual(state.current_player(), 0)
      state.apply_action(ACTION_CHECK_OR_CALL)
      # P1 BB calls all-in (has 80 behind, action is for 100)
      self.assertEqual(state.current_player(), 1)
      state.apply_action(ACTION_CHECK_OR_CALL)

      # Flop, turn, river
      board = [
          Card(TREY, HEART),
          Card(FOUR, DIAMOND),
          Card(SIX, SPADE),
          Card(SEVEN, HEART),
          Card(EIGHT, HEART),
      ]
      for card in board:
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[card])

      # Hand 0 finished, P0 won, P1 and P2 busted.
      # With only 1 player left, game should be terminal.
      self.assertTrue(state.is_terminal())
      self.assertEqual(state._hand_number, 1)
      self.assertEqual(state._stacks, [300, 0, 0])
      self.assertEqual(state._num_active_players, 1)

    def test_6max_3players_bust_then_split_pot(self):
      params = {
          "max_num_hands": 2,
          "reset_stacks": False,
          "rotate_dealer": True,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "num_players": 6,
              "blinds": "10 20",
              "stack_sizes": "100 100 100 100 100 100",
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()

      # Hand 0: 4-way all-in preflop, P0 wins, P1,P2,P3 bust.
      # 6 players: P0=SB, P1=BB, P2=UTG, P3=MP, P4=CO, P5=BTN
      self.assertEqual(state._hand_number, 0)
      self.assertEqual(state._stacks, [100] * 6)

      # Hole cards:
      # P0: Ac As
      # P1: Kc Ks
      # P2: Qc Qs
      # P3: Jc Js
      # P4: 2c 2s
      # P5: 3c 3s
      hole_cards_h0 = [
          Card(ACE, CLUB),
          Card(KING, CLUB),
          Card(QUEEN, CLUB),
          Card(JACK, CLUB),
          Card(DEUCE, CLUB),
          Card(TREY, CLUB),
          Card(ACE, SPADE),
          Card(KING, SPADE),
          Card(QUEEN, SPADE),
          Card(JACK, SPADE),
          Card(DEUCE, SPADE),
          Card(TREY, SPADE),
      ]
      for card in hole_cards_h0:
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[card])

      self.assertEqual(state.current_player(), 2)  # P2 UTG acts first preflop

      # P2 UTG raises all-in to 100
      state.apply_action(100)
      # P3 MP calls all-in
      self.assertEqual(state.current_player(), 3)
      state.apply_action(ACTION_CHECK_OR_CALL)
      # P4 CO folds
      self.assertEqual(state.current_player(), 4)
      state.apply_action(ACTION_FOLD)
      # P5 BTN folds
      self.assertEqual(state.current_player(), 5)
      state.apply_action(ACTION_FOLD)
      # P0 SB calls all-in
      self.assertEqual(state.current_player(), 0)
      state.apply_action(ACTION_CHECK_OR_CALL)
      # P1 BB calls all-in
      self.assertEqual(state.current_player(), 1)
      state.apply_action(ACTION_CHECK_OR_CALL)

      # Board cards
      board_h0 = [
          Card(SEVEN, DIAMOND),
          Card(SEVEN, SPADE),
          Card(SEVEN, CLUB),
          Card(EIGHT, DIAMOND),
          Card(EIGHT, HEART),
      ]
      for card in board_h0:
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[card])

      # Hand 0 finished. P0 won 300 chips. P1,P2,P3 busted.
      # Players 0,4,5 remain active.
      self.assertEqual(state._hand_number, 1)
      self.assertEqual(state._stacks, [400, 0, 0, 0, 100, 100])
      self.assertEqual(state._num_active_players, 3)
      self.assertFalse(state.is_terminal())

      # Hand 1: Players 0,4,5 play. 3-way split pot.
      # P0=BTN, P4=SB, P5=BB. Stacks: P0=400, P4=100, P5=100. Blinds 10/20.
      # Active players sorted are 0,4,5 -> player_map {0:0, 1:4, 2:5}, so
      # seat0=P0, seat1=P4, seat2=P5.
      # P0 is dealer, so P4 is SB, P5 is BB.
      # Hole cards for hand 1:
      # P0: 2h 3h, P4: 4s 5s, P5: 6c 7c
      hole_cards_h1 = [
          Card(DEUCE, HEART),  # P0 card 1
          Card(FOUR, SPADE),  # P4 card 1
          Card(SIX, CLUB),  # P5 card 1
          Card(TREY, HEART),  # P0 card 2
          Card(FIVE, SPADE),  # P4 card 2
          Card(SEVEN, CLUB),  # P5 card 2
      ]
      for card in hole_cards_h1:
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[card])

      # Players 0,4,5 are active. P0=BTN, P4=SB, P5=BB.
      # P0 P4 and P5 all raise as much as possible (or call) to force an all-in
      # situation.
      # In 3-handed, BTN acts first preflop.
      self.assertEqual(state.current_player(), 0)
      state.apply_action(max(state.legal_actions()))
      self.assertEqual(state.current_player(), 4)
      state.apply_action(max(state.legal_actions()))
      self.assertEqual(state.current_player(), 5)
      state.apply_action(max(state.legal_actions()))

      # Board Ad Kd Qd Jd Td - royal flush on board.
      board_h1 = [
          Card(ACE, DIAMOND),
          Card(KING, DIAMOND),
          Card(QUEEN, DIAMOND),
          Card(JACK, DIAMOND),
          Card(TEN, DIAMOND),
      ]
      for card in board_h1:
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.pokerkit_wrapper_game.card_to_int[card])

      # Hand 1 finished, split pot. Stacks unchanged.
      # max_num_hands will have been reached, so game should be terminal and
      # the hand number should remain the same as before.
      self.assertEqual(state._hand_number, 1)
      self.assertTrue(state.is_terminal())
      self.assertEqual(state._stacks, [400, 0, 0, 0, 100, 100])
      self.assertEqual(state._num_active_players, 3)
      self.assertEqual(
          state.returns(), [300.0, -100.0, -100.0, -100.0, 0.0, 0.0]
      )

    def test_max_num_hands_limit(self):
      """Tests that the game terminates when max_num_hands is reached."""
      params = {
          "max_num_hands": 1,
          "reset_stacks": False,
          "rotate_dealer": True,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "num_players": 2,
              "blinds": "50 100",
              "stack_sizes": "1000 1000",
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()
      self.assertEqual(state._hand_number, 0)
      self.assertFalse(state.is_terminal())

      # Play one hand: SB folds immediately preflop.
      # 4 chance nodes for hole cards
      for _ in range(4):
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.chance_outcomes()[0][0])

      self.assertFalse(state.is_chance_node())
      # P1 is SB and folds.
      self.assertEqual(state.current_player(), 1)
      state.apply_action(ACTION_FOLD)

      # Hand is over, max_num_hands=1 is reached, game should be terminal.
      self.assertTrue(state.is_terminal())
      self.assertEqual(state._hand_number, 0)
      # P1 is SB in hand 0 and folded. P0 is BB.
      # P1 stack: 1000 - 50 = 950. P0 stack: 1000 + 50 = 1050.
      self.assertEqual(state._stacks, [1050, 950])

    def test_returns_in_terminal_state(self):
      """Tests that returns() are correct when state is terminal."""
      params = {
          "max_num_hands": 1,
          "reset_stacks": False,
          "rotate_dealer": True,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "num_players": 2,
              "blinds": "50 100",
              "stack_sizes": "1000 1000",
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()
      self.assertEqual(state._hand_number, 0)
      self.assertFalse(state.is_terminal())

      # Play one hand: SB folds immediately preflop.
      # 4 chance nodes for hole cards
      for _ in range(4):
        self.assertTrue(state.is_chance_node())
        state.apply_action(state.chance_outcomes()[0][0])

      self.assertFalse(state.is_chance_node())
      # P1 is SB and folds.
      self.assertEqual(state.current_player(), 1)
      state.apply_action(ACTION_FOLD)

      # Hand is over, max_num_hands=1 is reached, game should be terminal.
      self.assertTrue(state.is_terminal())
      # P1 is SB in hand 0 and folded. P0 is BB.
      # P1 stack: 1000 - 50 = 950. P0 stack: 1000 + 50 = 1050.
      self.assertEqual(state.returns(), [50.0, -50.0])

    def test_record_full_hand_histories_true(self):
      params = {
          "max_num_hands": 20,
          "reset_stacks": False,
          "rotate_dealer": True,
          "record_full_hand_histories": True,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "num_players": 2,
              "blinds": "50 100",
              "stack_sizes": "1000 1000",
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()
      self.assertEqual(state._hand_number, 0)
      self.assertEmpty(state._wrapped_state_hand_histories)

      # Play a couple hands where SB folds immediately preflop.
      for _ in range(10):
        # 4 chance nodes for hole cards
        for _ in range(4):
          state.apply_action(state.chance_outcomes()[0][0])
        state.apply_action(ACTION_FOLD)

      # Crux of the test: we should have 10 hand histories recorded here.
      self.assertEqual(state._hand_number, 10)
      self.assertLen(state._wrapped_state_hand_histories, 10)
      for hand_history in state._wrapped_state_hand_histories:
        self.assertNotEmpty(hand_history)
      self.assertEqual(
          state._wrapped_state_hand_histories[0][0],
          # Expected p1 hand history, first hand (p2 hand is censored)
          ["d dh p1 2c2h", "d dh p2 ????", "p2 f"],
      )
      self.assertEqual(
          state._wrapped_state_hand_histories[0][1],
          # Expected p2 hand history, first hand (p1 hand is censored)
          ["d dh p1 ????", "d dh p2 2d2s", "p2 f"],
      )

    def test_record_full_hand_histories_false(self):
      params = {
          "max_num_hands": 20,
          "reset_stacks": False,
          "rotate_dealer": True,
          "record_full_hand_histories": False,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "num_players": 2,
              "blinds": "50 100",
              "stack_sizes": "1000 1000",
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()
      self.assertEqual(state._hand_number, 0)
      self.assertEmpty(state._wrapped_state_hand_histories)

      # Play a couple hands where SB folds immediately preflop.
      for _ in range(10):
        # 4 chance nodes for hole cards
        for _ in range(4):
          state.apply_action(state.chance_outcomes()[0][0])
        state.apply_action(ACTION_FOLD)

      # Crux of the test: still no hand histories recorded here.
      self.assertEqual(state._hand_number, 10)
      self.assertEmpty(state._wrapped_state_hand_histories)

    def test_pokerkit_state_to_struct_to_json_two_hands(self):
      params = {
          "max_num_hands": 2,
          "reset_stacks": False,
          "rotate_dealer": True,
          "pokerkit_game_params": {
              "name": "python_pokerkit_wrapper",
              "num_players": 2,
              "blinds": "50 100",
              "stack_sizes": "1000 1000",
          },
      }
      game = pyspiel.load_game("python_repeated_pokerkit", params)
      state = game.new_initial_state()
      self.assertEqual(state._hand_number, 0)

      state_struct = state.to_struct()
      state_json = state.to_json()
      self.assertEqual(state_struct.to_json(), state_json)

      # --- First hand (hand 0) ---
      self.assertIsInstance(state_json, str)
      try:
        data_h0 = json.loads(state_json)
      except json.JSONDecodeError as e:
        self.fail(f"Failed to parse JSON {state_json}, error: {e}")

      self.assertEqual(state_struct.hand_number, 0)
      self.assertEqual(data_h0["hand_number"], 0)
      self.assertEqual(state_struct.is_terminal, False)
      self.assertEqual(data_h0["is_terminal"], False)
      self.assertEqual(state_struct.stacks, [1000, 1000])
      self.assertEqual(data_h0["stacks"], [1000, 1000])
      self.assertEqual(state_struct.dealer, 1)
      self.assertEqual(data_h0["dealer"], 1)
      self.assertEqual(state_struct.seat_to_player, {0: 0, 1: 1})
      self.assertEqual(data_h0["seat_to_player"], [[0, 0], [1, 1]])
      self.assertEqual(state_struct.player_to_seat, {0: 0, 1: 1})
      self.assertEqual(data_h0["player_to_seat"], [[0, 0], [1, 1]])
      self.assertEqual(state_struct.small_blind, 50)
      self.assertEqual(data_h0["small_blind"], 50)
      self.assertEqual(state_struct.big_blind, 100)
      self.assertEqual(data_h0["big_blind"], 100)
      self.assertEqual(state_struct.small_bet_size, -1)
      self.assertEqual(data_h0["small_bet_size"], -1)
      self.assertEqual(state_struct.big_bet_size, -1)
      self.assertEqual(data_h0["big_bet_size"], -1)
      self.assertEqual(state_struct.bring_in, -1)
      self.assertEqual(data_h0["bring_in"], -1)
      self.assertEqual(state_struct.hand_returns, [[0.0, 0.0]])
      self.assertEqual(data_h0["hand_returns"], [[0.0, 0.0]])

      pk_state_struct_h0 = state_struct.pokerkit_state_struct
      pk_data_h0 = data_h0["pokerkit_state_struct"]

      self.assertEqual(pk_data_h0["is_terminal"], False)
      self.assertEqual(
          state.pokerkit_wrapper_state.is_terminal(), pk_data_h0["is_terminal"]
      )
      self.assertEqual(pk_data_h0["current_player"], pyspiel.PlayerId.CHANCE)
      self.assertEqual(
          state.pokerkit_wrapper_state.current_player(),
          pk_data_h0["current_player"],
      )
      # Stacks in pokerkit state are after blinds: 1000-50=950, 1000-100=900
      self.assertEqual(pk_data_h0["stacks"], [900, 950])
      self.assertEqual(pk_state_struct_h0.stacks, pk_data_h0["stacks"])
      self.assertEqual(pk_data_h0["bets"], [100, 50])
      self.assertEqual(pk_state_struct_h0.bets, pk_data_h0["bets"])

      # Play until hand is over, having P1 fold immediately preflop.
      while state._hand_number == 0:
        if state.is_chance_node():
          state.apply_action(state.chance_outcomes()[0][0])
        else:
          state.apply_action(ACTION_FOLD)

      # --- Second hand (hand 1) ---
      self.assertEqual(state._hand_number, 1)
      state_struct_h1 = state.to_struct()
      state_json_h1 = state.to_json()
      self.assertEqual(state_struct_h1.to_json(), state_json_h1)
      self.assertIsInstance(state_json_h1, str)
      try:
        data_h1 = json.loads(state_json_h1)
      except json.JSONDecodeError as e:
        self.fail(f"Failed to parse JSON {state_json_h1}, error: {e}")

      self.assertEqual(state_struct_h1.hand_number, 1)
      self.assertEqual(data_h1["hand_number"], 1)
      self.assertEqual(state_struct_h1.is_terminal, False)
      self.assertEqual(data_h1["is_terminal"], False)
      # P1 folded in hand 0, so P0 wins 50 from P1.
      # Stacks before hand 1: P0=1000+50=1050, P1=1000-50=950.
      self.assertEqual(state_struct_h1.stacks, [1050, 950])
      self.assertEqual(data_h1["stacks"], [1050, 950])
      self.assertEqual(state_struct_h1.dealer, 0)
      self.assertEqual(data_h1["dealer"], 0)
      self.assertEqual(state_struct_h1.seat_to_player, {0: 1, 1: 0})
      self.assertEqual(data_h1["seat_to_player"], [[0, 1], [1, 0]])
      self.assertEqual(state_struct_h1.player_to_seat, {0: 1, 1: 0})
      self.assertEqual(data_h1["player_to_seat"], [[0, 1], [1, 0]])
      self.assertEqual(state_struct_h1.small_blind, 50)
      self.assertEqual(data_h1["small_blind"], 50)
      self.assertEqual(state_struct_h1.big_blind, 100)
      self.assertEqual(data_h1["big_blind"], 100)
      self.assertEqual(state_struct_h1.small_bet_size, -1)
      self.assertEqual(data_h1["small_bet_size"], -1)
      self.assertEqual(state_struct_h1.big_bet_size, -1)
      self.assertEqual(data_h1["big_bet_size"], -1)
      self.assertEqual(state_struct_h1.bring_in, -1)
      self.assertEqual(data_h1["bring_in"], -1)
      self.assertEqual(
          state_struct_h1.hand_returns, [[50.0, -50.0], [0.0, 0.0]]
      )
      self.assertEqual(data_h1["hand_returns"], [[50.0, -50.0], [0.0, 0.0]])

      pk_state_struct_h1 = state_struct_h1.pokerkit_state_struct
      pk_data_h1 = data_h1["pokerkit_state_struct"]

      self.assertEqual(pk_data_h1["is_terminal"], False)
      self.assertEqual(
          pk_state_struct_h1.is_terminal, pk_data_h1["is_terminal"]
      )
      self.assertEqual(pk_data_h1["current_player"], pyspiel.PlayerId.CHANCE)
      self.assertEqual(
          pk_state_struct_h1.current_player, pk_data_h1["current_player"]
      )
      # Dealer rotates, P1 was dealer in hand 0, so P0 is dealer in hand 1.
      # Heads-up: dealer is SB. P0=SB=50, P1=BB=100.
      # Pokerkit state stacks based on seats:
      # P0 is seat 1 (SB), stack 1050-50=1000
      # P1 is seat 0 (BB), stack 950-100=850
      self.assertEqual(pk_data_h1["stacks"], [850, 1000])
      self.assertEqual(pk_state_struct_h1.stacks, pk_data_h1["stacks"])
      self.assertEqual(pk_data_h1["bets"], [100, 50])
      self.assertEqual(pk_state_struct_h1.bets, pk_data_h1["bets"])

      # Finally double check we can reach terminal state as expected
      while not state.is_terminal():
        if state.is_chance_node():
          state.apply_action(state.chance_outcomes()[0][0])
        else:
          # Now this time the other player folds - meaning we're inverting what
          # just happened last hand.
          state.apply_action(ACTION_FOLD)

      state_struct_terminal = state.to_struct()
      state_json_terminal = state.to_json()
      self.assertIsInstance(state_json_terminal, str)
      try:
        data_terminal = json.loads(state_json_terminal)
      except json.JSONDecodeError as e:
        self.fail(f"Failed to parse JSON {state_json_terminal}, error: {e}")
      self.assertTrue(data_terminal["is_terminal"], True)
      self.assertEqual(
          state_struct_terminal.is_terminal, data_terminal["is_terminal"]
      )
      self.assertEqual(
          state_struct_terminal.hand_returns, [[50.0, -50.0], [-50.0, 50.0]]
      )
      self.assertEqual(
          data_terminal["hand_returns"], [[50.0, -50.0], [-50.0, 50.0]]
      )
      self.assertEqual(state_struct_terminal.stacks, [1000, 1000])
      self.assertEqual(data_terminal["stacks"], [1000, 1000])

      # Max hands was 2, so it shouldn't have progressed past hand 1 and the
      # players shouldn't have rotated.
      self.assertEqual(state_struct_terminal.hand_number, 1)
      self.assertEqual(data_terminal["hand_number"], 1)
      self.assertEqual(state_struct_terminal.seat_to_player, {0: 1, 1: 0})
      self.assertEqual(data_terminal["seat_to_player"], [[0, 1], [1, 0]])
      self.assertEqual(state_struct_terminal.player_to_seat, {0: 1, 1: 0})
      self.assertEqual(data_terminal["player_to_seat"], [[0, 1], [1, 0]])


if __name__ == "__main__":
  if IMPORTED_ALL_LIBRARIES:
    absltest.main()
  else:
    logging.warning("Skipping test because not all libraries were imported.")



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for basic heuristics."""

import functools
from absl.testing import absltest
import numpy as np
from open_spiel.python.algorithms import minimax
from open_spiel.python.games import heuristics
import pyspiel


def simulate_random_game(game: pyspiel.Game):
  """Simulates a game."""
  state = game.new_initial_state()
  while not state.is_terminal():
    # The state can be three different types: chance node,
    # simultaneous node, or decision node
    if state.is_chance_node():
      outcomes = state.chance_outcomes()
      action_list, prob_list = zip(*outcomes)
      action = np.random.choice(action_list, p=prob_list)
      state.apply_action(action)
    elif state.is_simultaneous_node():
      heuristic_value = heuristics.evaluate_state(state, state.current_player())
      print(f"---\nState is: \n{state}\nHeuristic value: {heuristic_value}\n")
      random_choice = lambda a: np.random.choice(a) if a else [0]
      chosen_actions = [
          random_choice(state.legal_actions(pid))
          for pid in range(game.num_players())
      ]
      state.apply_actions(chosen_actions)
    else:
      heuristic_value = heuristics.evaluate_state(state, state.current_player())
      print(f"---\nState is: \n{state}\nHeuristic value: {heuristic_value}\n")
      action = np.random.choice(state.legal_actions(state.current_player()))
      state.apply_action(action)


class HeuristicsTest(absltest.TestCase):

  def test_can_call_heuristic_callback(self):
    """Checks we can create the game and a state."""
    for game_str in heuristics.HEURISTIC_CALLBACKS:
      game = pyspiel.load_game(game_str)
      simulate_random_game(game)

  def test_minimax_chess(self):
    """Tests minimax for chess."""
    game = pyspiel.load_game("chess")
    state = game.new_initial_state()
    while not state.is_terminal():
      player = state.current_player()
      if player == 0:
        action = np.random.choice(state.legal_actions(state.current_player()))
      else:
        _, action = minimax.alpha_beta_search(
            game=game,
            state=state,
            value_function=functools.partial(heuristics.evaluate_state,
                                             player=player),
            maximum_depth=2,
            maximizing_player_id=player)
      print(f"State is: \n{state.debug_string()}\n")
      print(f"Taking action: {state.action_to_string(player, action)}\n")
      state.apply_action(action)


if __name__ == "__main__":
  absltest.main()
